# -*- coding: utf-8 -*-
import pytest
import unittest

import numpy as np
import astropy.wcs

from ndcube import utils
from ndcube.tests import helpers


ht = {'CTYPE3': 'HPLT-TAN', 'CUNIT3': 'deg', 'CDELT3': 0.5, 'CRPIX3': 0, 'CRVAL3': 0, 'NAXIS3': 2,
      'CTYPE2': 'WAVE    ', 'CUNIT2': 'Angstrom', 'CDELT2': 0.2, 'CRPIX2': 0, 'CRVAL2': 0,
      'NAXIS2': 3,
      'CTYPE1': 'TIME    ', 'CUNIT1': 'min', 'CDELT1': 0.4, 'CRPIX1': 0, 'CRVAL1': 0, 'NAXIS1': 4}
wt = utils.wcs.WCS(header=ht, naxis=3)

ht_with_celestial = {
    'CTYPE4': 'HPLN-TAN', 'CUNIT4': 'deg', 'CDELT4': 1, 'CRPIX4': 0, 'CRVAL4': 0, 'NAXIS4': 1,
    'CNAME4': 'redundant axis', 'CROTA4': 0,
    'CTYPE3': 'HPLT-TAN', 'CUNIT3': 'deg', 'CDELT3': 0.5, 'CRPIX3': 0, 'CRVAL3': 0, 'NAXIS3': 2,
    'CTYPE2': 'WAVE    ', 'CUNIT2': 'Angstrom', 'CDELT2': 0.2, 'CRPIX2': 0, 'CRVAL2': 0,
    'NAXIS2': 3,
    'CTYPE1': 'TIME    ', 'CUNIT1': 'min', 'CDELT1': 0.4, 'CRPIX1': 0, 'CRVAL1': 0, 'NAXIS1': 4}

hm = {'CTYPE1': 'WAVE    ', 'CUNIT1': 'Angstrom', 'CDELT1': 0.2, 'CRPIX1': 0, 'CRVAL1': 10,
      'NAXIS1': 4,
      'CTYPE2': 'HPLT-TAN', 'CUNIT2': 'deg', 'CDELT2': 0.5, 'CRPIX2': 2, 'CRVAL2': 0.5,
      'NAXIS2': 3,
      'CTYPE3': 'HPLN-TAN', 'CUNIT3': 'deg', 'CDELT3': 0.4, 'CRPIX3': 2, 'CRVAL3': 1, 'NAXIS3': 2}
wm = utils.wcs.WCS(header=hm, naxis=3)

hm_reindexed_102 = {
    'CTYPE2': 'WAVE    ', 'CUNIT2': 'Angstrom', 'CDELT2': 0.2, 'CRPIX2': 0, 'CRVAL2': 10,
    'NAXIS2': 4,
    'CTYPE1': 'HPLT-TAN', 'CUNIT1': 'deg', 'CDELT1': 0.5, 'CRPIX1': 2, 'CRVAL1': 0.5, 'NAXIS1': 3,
    'CTYPE3': 'HPLN-TAN', 'CUNIT3': 'deg', 'CDELT3': 0.4, 'CRPIX3': 2, 'CRVAL3': 1, 'NAXIS3': 2}
wm_reindexed_102 = utils.wcs.WCS(header=hm_reindexed_102, naxis=3)


@pytest.mark.parametrize("test_input,expected", [(ht, True), (hm, False)])
def test_wcs_needs_augmenting(test_input, expected):
    assert utils.wcs.WCS._needs_augmenting(test_input) is expected


@pytest.mark.parametrize("test_input,expected", [((ht, 3), ht_with_celestial)])
def test_wcs_augment(test_input, expected):
    unit_tester = unittest.TestCase()
    unit_tester.assertEqual(utils.wcs.WCS._augment(*test_input), expected)


@pytest.mark.parametrize(
    "test_input,expected",
    [({}, False),
     ([slice(1, 5), slice(-1, -5, -2)], True)])
def test_all_slice(test_input, expected):
    assert utils.wcs._all_slice(test_input) == expected


@pytest.mark.parametrize(
    "test_input,expected",
    [({}, []),
     ((slice(1, 2), slice(1, 3), 2, slice(2, 4), 8),
      [slice(1, 2, None), slice(1, 3, None), slice(2, 3, None),
       slice(2, 4, None), slice(8, 9, None)])])
def test_slice_list(test_input, expected):
    assert utils.wcs._slice_list(test_input) == expected


@pytest.mark.parametrize("test_input,expected", [
    ((wm, np.array([1, 0, 2])), wm_reindexed_102),
    ((wm, np.array([1, 0, -1])), wm_reindexed_102)
    ])
def test_reindex_wcs(test_input, expected):
    print(utils.wcs.reindex_wcs(*test_input))
    print(expected)
    helpers.assert_wcs_are_equal(utils.wcs.reindex_wcs(*test_input), expected)


@pytest.mark.parametrize("test_input", [
    (TypeError, wm, 0),
    (TypeError, wm, np.array(['spam', 'eggs', 'ham'])),
    ])
def test_reindex_wcs_errors(test_input):
    with pytest.raises(test_input[0]):
        utils.wcs.reindex_wcs(*test_input[1:])


@pytest.mark.parametrize("test_input,expected", [
    ((wm, 0, [False, False, False]), (0, 1)),
    ((wm, 1, [False, False, False]), (0, 1)),
    ((wm, 2, [False, False, False]), (2,)),
    ((wm, 1, [False, False, True]), (1,))
    ])
def test_get_dependent_data_axes(test_input, expected):
    output = utils.wcs.get_dependent_data_axes(*test_input)
    assert output == expected


@pytest.mark.parametrize("test_input,expected", [
    ((wm, 0), (0,)),
    ((wm, 1), (1, 2)),
    ((wm, 2), (1, 2)),
    ])
def test_get_dependent_wcs_axes(test_input, expected):
    output = utils.wcs.get_dependent_wcs_axes(*test_input)
    assert output == expected


@pytest.mark.parametrize("test_input,expected", [
    (wm, np.array([[True, False, False], [False, True, True], [False, True, True]])),
    (wt, np.array([[True, False, False, False], [False, True, False, False],
                   [False, False, True, True], [False, False, True, True]])),
    (wm_reindexed_102, np.array([[True, False, True], [False, True, False],
                                 [True, False, True]]))
    ])
def test_axis_correlation_matrix(test_input, expected):
    assert (utils.wcs.axis_correlation_matrix(test_input) == expected).all()



# -*- coding: utf-8 -*-
import pytest
import unittest

import numpy as np

from ndcube import utils

# sample data for tests
tuple_item0 = (0, slice(0, 3))
tuple_item1 = (slice(0, 2), slice(0, 3), slice(None))
tuple_item2 = (slice(3, 1, -1), slice(0, 3), slice(None))
tuple_item3 = (slice(4, None, -2), slice(0, 3), slice(None))

n_cubes = 4


@pytest.mark.parametrize("test_input,expected", [
    ((1, n_cubes), [utils.sequence.SequenceItem(sequence_index=1, cube_item=slice(None))]),
    ((slice(None), 2), [utils.sequence.SequenceItem(sequence_index=0, cube_item=slice(None)),
                        utils.sequence.SequenceItem(sequence_index=1, cube_item=slice(None))]),
    ((slice(0, 2), 3), [utils.sequence.SequenceItem(sequence_index=0, cube_item=slice(None)),
                        utils.sequence.SequenceItem(sequence_index=1, cube_item=slice(None))]),
    ((slice(1, 4, 2), 5), [utils.sequence.SequenceItem(sequence_index=1, cube_item=slice(None)),
                           utils.sequence.SequenceItem(sequence_index=3, cube_item=slice(None))]),
    ((slice(3, 1, -1), 5), [utils.sequence.SequenceItem(sequence_index=3, cube_item=slice(None)),
                            utils.sequence.SequenceItem(sequence_index=2, cube_item=slice(None))]),
    ((tuple_item0, n_cubes),
     [utils.sequence.SequenceItem(sequence_index=0, cube_item=tuple_item0[1])]),
    ((tuple_item1, n_cubes),
     [utils.sequence.SequenceItem(sequence_index=0, cube_item=tuple_item1[1:]),
      utils.sequence.SequenceItem(sequence_index=1, cube_item=tuple_item1[1:])]),
    ((tuple_item2, n_cubes),
     [utils.sequence.SequenceItem(sequence_index=3, cube_item=tuple_item1[1:]),
      utils.sequence.SequenceItem(sequence_index=2, cube_item=tuple_item1[1:])]),
    ((tuple_item3, n_cubes),
     [utils.sequence.SequenceItem(sequence_index=4, cube_item=tuple_item1[1:]),
      utils.sequence.SequenceItem(sequence_index=2, cube_item=tuple_item1[1:]),
      utils.sequence.SequenceItem(sequence_index=0, cube_item=tuple_item1[1:])])
    ])
def test_convert_item_to_sequence_items(test_input, expected):
    unit_tester = unittest.TestCase()
    unit_tester.assertEqual(
        utils.sequence.convert_item_to_sequence_items(*test_input), expected)


def test_convert_item_to_sequence_items_error():
    with pytest.raises(TypeError):
        utils.sequence.convert_item_to_sequence_items('item')


@pytest.mark.parametrize("test_input,expected", [
    # Test int cube_like_items.
    ((0, 0, np.array([3])), [utils.sequence.SequenceItem(sequence_index=0, cube_item=0)]),
    ((5, 0, np.array([3, 3])), [utils.sequence.SequenceItem(sequence_index=1, cube_item=2)]),
    # Below test reveals function doesn't work with negative int indexing.
    #((-1, 0, np.array([3, 3])), [utils.sequence.SequenceItem(sequence_index=1, cube_item=2)]),

    # Test slice cube_like_items.
    ((slice(0, 2), 0, np.array([3])),
     [utils.sequence.SequenceItem(sequence_index=0, cube_item=slice(0, 2, 1))]),
    ((slice(1, 4), 0, np.array([3, 3])),
     [utils.sequence.SequenceItem(sequence_index=0, cube_item=slice(1, 3, 1)),
      utils.sequence.SequenceItem(sequence_index=1, cube_item=slice(0, 1, 1))]),
    ((slice(1, 7, 2), 0, np.array([3, 5])),
     [utils.sequence.SequenceItem(sequence_index=0, cube_item=slice(1, 3, 2)),
      utils.sequence.SequenceItem(sequence_index=1, cube_item=slice(1, 4, 2))]),
    ((slice(1, 7, 3), 0, np.array([3, 5])),
     [utils.sequence.SequenceItem(sequence_index=0, cube_item=slice(1, 2, 3)),
      utils.sequence.SequenceItem(sequence_index=1, cube_item=slice(1, 4, 3))]),
    # Below test reveals function doesn't work with negative stepping.
    #((slice(6, 1, -1), 0, np.array([3, 5])),
    # [utils.sequence.SequenceItem(sequence_index=1, cube_item=slice(3, 0, -1)),
    #  utils.sequence.SequenceItem(sequence_index=0, cube_item=slice(2, 1, -1))]),

    # Test tuple cube_like_items
    (((0, 0, slice(1, 10)), 0, np.array([3, 5])),
     [utils.sequence.SequenceItem(sequence_index=0, cube_item=(0, 0, slice(1, 10, None)))]),

    (((0, 0, slice(1, 10)), 1, np.array([3, 5])),
     [utils.sequence.SequenceItem(sequence_index=0, cube_item=(0, 0, slice(1, 10, None)))]),

    (((slice(2, 10), 0, slice(1, 10)), 0, np.array([3, 5, 5])),
     [utils.sequence.SequenceItem(sequence_index=0,
                                  cube_item=(slice(2, 3, 1), 0, slice(1, 10, None))),
      utils.sequence.SequenceItem(sequence_index=1,
                                  cube_item=(slice(0, 5, 1), 0, slice(1, 10, None))),
      utils.sequence.SequenceItem(sequence_index=2,
                                  cube_item=(slice(0, 2, 1), 0, slice(1, 10, None)))]),

    (((0, slice(2, 10), slice(1, 10)), 1, np.array([3, 5, 5])),
     [utils.sequence.SequenceItem(sequence_index=0,
                                  cube_item=(0, slice(2, 3, 1), slice(1, 10, None))),
      utils.sequence.SequenceItem(sequence_index=1,
                                  cube_item=(0, slice(0, 5, 1), slice(1, 10, None))),
      utils.sequence.SequenceItem(sequence_index=2,
                                  cube_item=(0, slice(0, 2, 1), slice(1, 10, None)))]),
     ])
def test_convert_cube_like_item_to_sequence_items(test_input, expected):
    unit_tester = unittest.TestCase()
    unit_tester.assertEqual(
        utils.sequence.convert_cube_like_item_to_sequence_items(*test_input), expected)


@pytest.mark.parametrize("test_input", [
    (0, 1, np.array(3)), (slice(None), 1, np.array(3)),
    ((0, 1), 2, np.array([3, 3, 3])), (('item', 2), 0, np.array([3, 3, 3]))
    ])
def test_convert_cube_like_item_to_sequence_items_value_error(test_input):
    with pytest.raises(ValueError):
        utils.sequence.convert_cube_like_item_to_sequence_items(*test_input)


def test_convert_cube_like_item_to_sequence_items_type_error():
    with pytest.raises(TypeError):
        utils.sequence.convert_cube_like_item_to_sequence_items('item', 1, np.array(3))


@pytest.mark.parametrize("test_input,expected", [
    ((5, np.array([8] * 4)), utils.sequence.SequenceSlice(0, 5)),
    ((8, np.array([8] * 4)), utils.sequence.SequenceSlice(1, 0)),
    ((20, np.array([8] * 4)), utils.sequence.SequenceSlice(2, 4)),
    ((50, np.array([8] * 4)), utils.sequence.SequenceSlice(3, 8)),
])
def test_convert_cube_like_index_to_sequence_slice(test_input, expected):
    assert utils.sequence._convert_cube_like_index_to_sequence_slice(
        *test_input) == expected


@pytest.mark.parametrize("test_input,expected",
                         [((slice(2, 5), np.array([8] * 4)),
                           [utils.sequence.SequenceSlice(0, slice(2, 5, 1))]),
                          ((slice(5, 15), np.array([8] * 4)), [
                              utils.sequence.SequenceSlice(0, slice(5, 8, 1)),
                              utils.sequence.SequenceSlice(1, slice(0, 7, 1))
                          ]), ((slice(5, 16), np.array([8] * 4)), [
                              utils.sequence.SequenceSlice(0, slice(5, 8, 1)),
                              utils.sequence.SequenceSlice(1, slice(0, 8, 1))
                          ]), ((slice(5, 23), np.array([8] * 4)), [
                              utils.sequence.SequenceSlice(0, slice(5, 8, 1)),
                              utils.sequence.SequenceSlice(1, slice(0, 8, 1)),
                              utils.sequence.SequenceSlice(2, slice(0, 7, 1))
                          ]), ((slice(5, 100), np.array([8] * 4)), [
                              utils.sequence.SequenceSlice(0, slice(5, 8, 1)),
                              utils.sequence.SequenceSlice(1, slice(0, 8, 1)),
                              utils.sequence.SequenceSlice(2, slice(0, 8, 1)),
                              utils.sequence.SequenceSlice(3, slice(0, 8, 1))
                          ])])
def test_convert_cube_like_slice_to_sequence_slices(test_input, expected):
    assert utils.sequence._convert_cube_like_slice_to_sequence_slices(*test_input) == expected


@pytest.mark.parametrize(
    "test_input,expected",
    [((slice(0, 10), 20), slice(0, 10, 1)),
     ((slice(0, 10, 2), 20), slice(0, 10, 2)),
     ((slice(None, 0, -1), 20), slice(20, 0, -1))])
def test_convert_slice_nones_to_ints(test_input, expected):
    assert utils.sequence.convert_slice_nones_to_ints(*test_input) == expected


@pytest.mark.parametrize("test_input,expected", [
    ((utils.sequence.SequenceSlice(0, 0), 1), utils.sequence.SequenceItem(0, (slice(None), 0)))
    ])
def test_convert_sequence_slice_to_sequence_item(test_input, expected):
    unit_tester = unittest.TestCase()
    unit_tester.assertEqual(
        utils.sequence._convert_sequence_slice_to_sequence_item(*test_input), expected)



import numpy as np

import astropy.units as u

from ndcube import utils
from ndcube.mixins.sequence_plotting import NDCubeSequencePlotMixin

__all__ = ['NDCubeSequence']


class NDCubeSequenceBase:
    """
    Class representing list of cubes.

    Parameters
    ----------
    data_list : `list`
        List of cubes.

    meta : `dict` or None
        The header of the NDCubeSequence.

    common_axis: `int` or None
        The data axis which is common between the NDCubeSequence and the Cubes within.
        For example, if the Cubes are sequenced in chronological order and time is
        one of the zeroth axis of each Cube, then common_axis should be se to 0.
        This enables the option for the NDCubeSequence to be indexed as though it is
        one single Cube.
    """

    def __init__(self, data_list, meta=None, common_axis=None, **kwargs):
        self.data = data_list
        self.meta = meta
        if common_axis is not None:
            self._common_axis = int(common_axis)
        else:
            self._common_axis = common_axis

    @property
    def dimensions(self):
        return self._dimensions

    @property
    def _dimensions(self):
        dimensions = [len(self.data) * u.pix] + list(self.data[0].dimensions)
        if len(dimensions) > 1:
            # If there is a common axis, length of cube's along it may not
            # be the same. Therefore if the lengths are different,
            # represent them as a tuple of all the values, else as an int.
            if self._common_axis is not None:
                common_axis_lengths = [cube.data.shape[self._common_axis] for cube in self.data]
                if len(np.unique(common_axis_lengths)) != 1:
                    common_axis_dimensions = [cube.dimensions[self._common_axis]
                                              for cube in self.data]
                    dimensions[self._common_axis+1] = u.Quantity(
                        common_axis_dimensions, unit=common_axis_dimensions[0].unit)
        return tuple(dimensions)

    @property
    def world_axis_physical_types(self):
        return tuple(["meta.obs.sequence"]+list(self.data[0].world_axis_physical_types))

    @property
    def cube_like_dimensions(self):
        if type(self._common_axis) is not int:
            raise TypeError("Common axis must be set.")
        dimensions = list(self._dimensions)
        cube_like_dimensions = list(self._dimensions[1:])
        if dimensions[self._common_axis+1].isscalar:
            cube_like_dimensions[self._common_axis] = \
              u.Quantity(dimensions[0].value * dimensions[self._common_axis+1].value, unit=u.pix)
        else:
            cube_like_dimensions[self._common_axis] = sum(dimensions[self._common_axis+1])
        # Combine into single Quantity
        cube_like_dimensions = u.Quantity(cube_like_dimensions, unit=u.pix)
        return cube_like_dimensions

    @property
    def cube_like_world_axis_physical_types(self):
        return self.data[0].world_axis_physical_types

    def __getitem__(self, item):
        if len(self.dimensions) == 1:
            return self.data[item]
        else:
            return utils.sequence.slice_sequence(self, item)

    @property
    def index_as_cube(self):
        """
        Method to slice the NDCubesequence instance as a single cube

        Example
        -------
        >>> # Say we have three Cubes each cube has common_axis=0 is time and shape=(3,3,3)
        >>> data_list = [cubeA, cubeB, cubeC] # doctest: +SKIP
        >>> cs = NDCubeSequence(data_list, meta=None, common_axis=0) # doctest: +SKIP
        >>> # return zeroth time slice of cubeB in via normal NDCubeSequence indexing.
        >>> cs[1,:,0,:] # doctest: +SKIP
        >>> # Return same slice using this function
        >>> cs.index_sequence_as_cube[3:6, 0, :] # doctest: +SKIP
        """
        if self._common_axis is None:
            raise ValueError("common_axis cannot be None")
        return _IndexAsCubeSlicer(self)

    @property
    def common_axis_extra_coords(self):
        if not isinstance(self._common_axis, int):
            raise ValueError("Common axis is not set.")
        # Get names and units of coords along common axis.
        axis_coord_names, axis_coord_units = utils.sequence._get_axis_extra_coord_names_and_units(
            self.data, self._common_axis)
        # Compile dictionary of common axis extra coords.
        if axis_coord_names is not None:
            return utils.sequence._get_int_axis_extra_coords(
                self.data, axis_coord_names, axis_coord_units, self._common_axis)
        else:
            return None

    @property
    def sequence_axis_extra_coords(self):
        sequence_coord_names, sequence_coord_units = \
          utils.sequence._get_axis_extra_coord_names_and_units(self.data, None)
        if sequence_coord_names is not None:
            # Define empty dictionary which will hold the extra coord
            # values not assigned a cube data axis.
            sequence_extra_coords = {}
            # Define list of None signifying unit of each coord.  It will
            # be filled in in for loop below.
            sequence_coord_units = [None]*len(sequence_coord_names)
            # Iterate through cubes and populate values of each extra coord
            # not assigned a cube data axis.
            cube_extra_coords = [cube.extra_coords for cube in self.data]
            for i, coord_key in enumerate(sequence_coord_names):
                coord_values = np.array([None]*len(self.data), dtype=object)
                for j, cube in enumerate(self.data):
                    # Construct list of coord values from each cube for given extra coord.
                    try:
                        coord_values[j] = cube_extra_coords[j][coord_key]["value"]
                        # Determine whether extra coord is a quantity by checking
                        # whether any one value has a unit. As we are not
                        # assuming that all cubes have the same extra coords
                        # along the sequence axis, we will keep checking as we
                        # move through the cubes until all cubes are checked or
                        # we have found a unit.
                        if (isinstance(cube_extra_coords[j][coord_key]["value"], u.Quantity) and
                                not sequence_coord_units[i]):
                            sequence_coord_units[i] = cube_extra_coords[j][coord_key]["value"].unit
                    except KeyError:
                        pass
                # If the extra coord is normally a Quantity, replace all
                # None occurrences in coord value array with a NaN, and
                # convert coord_values from an array of Quantities to a
                # single Quantity of length equal to number of cubes in
                # sequence.
                w_none = np.where(coord_values == None)[0]
                if sequence_coord_units[i]:
                    # This part of if statement is coded in an apparently
                    # round about way but necessitated because you can't
                    # put a NaN quantity into an array and keep its unit.
                    w_not_none = np.where(coord_values != None)[0]
                    coord_values = u.Quantity(list(coord_values[w_not_none]),
                                              unit=sequence_coord_units[i])
                    coord_values = list(coord_values.value)
                    for index in w_none:
                        coord_values.insert(index, np.nan)
                    coord_values = u.Quantity(coord_values, unit=sequence_coord_units[i]).flatten()
                else:
                    coord_values[w_none] = np.nan
                sequence_extra_coords[coord_key] = coord_values
        else:
            sequence_extra_coords = None
        return sequence_extra_coords

    def explode_along_axis(self, axis):
        """
        Separates slices of NDCubes in sequence along a given cube axis into (N-1)DCubes.

        Parameters
        ----------

        axis : `int`
            The axis along which the data is to be changed.
        """
        # if axis is None then set axis as common axis.
        if self._common_axis is not None:
            if self._common_axis != axis:
                raise ValueError("axis and common_axis should be equal.")
        # is axis is -ve then calculate the axis from the length of the dimensions of one cube
        if axis < 0:
            axis = len(self.dimensions[1::]) + axis
        # To store the resultant cube
        result_cubes = []
        # All slices are initially initialised as slice(None, None, None)
        result_cubes_slice = [slice(None, None, None)] * len(self[0].data.shape)
        # the range of the axis that needs to be sliced
        range_of_axis = self[0].data.shape[axis]
        for ndcube in self.data:
            for index in range(range_of_axis):
                # setting the slice value to the index so that the slices are done correctly.
                result_cubes_slice[axis] = index
                # appending the sliced cubes in the result_cube list
                result_cubes.append(ndcube.__getitem__(tuple(result_cubes_slice)))
        # creating a new sequence with the result_cubes keeping the meta and common axis as axis
        return self._new_instance(result_cubes, meta=self.meta)

    def __repr__(self):
        return (
            """NDCubeSequence
---------------------
Length of NDCubeSequence:  {length}
Shape of 1st NDCube: {shapeNDCube}
Axis Types of 1st NDCube: {axis_type}
""".format(length=self.dimensions[0], shapeNDCube=self.dimensions[1::],
           axis_type=self.world_axis_physical_types[1:]))

    @classmethod
    def _new_instance(cls, data_list, meta=None, common_axis=None):
        """
        Instantiate a new instance of this class using given data.
        """
        return cls(data_list, meta=meta, common_axis=common_axis)


class NDCubeSequence(NDCubeSequenceBase, NDCubeSequencePlotMixin):
    pass


"""
Cube Sequence Helpers
"""


class _IndexAsCubeSlicer:
    """
    Helper class to make slicing in index_as_cube sliceable/indexable
    like a numpy array.

    Parameters
    ----------
    seq : `ndcube.NDCubeSequence`
        Object of NDCubeSequence.

    """

    def __init__(self, seq):
        self.seq = seq

    def __getitem__(self, item):
        return utils.sequence._index_sequence_as_cube(self.seq, item)



# -*- coding: utf-8 -*-
from collections import namedtuple
import pytest
import datetime
import unittest

import sunpy.map
import numpy as np
import astropy.units as u

from ndcube import NDCube, NDCubeSequence
from ndcube.utils.wcs import WCS


# sample data for tests
# TODO: use a fixture reading from a test file. file TBD.
data = np.array([[[1, 2, 3, 4], [2, 4, 5, 3], [0, -1, 2, 3]],
                 [[2, 4, 5, 1], [10, 5, 2, 2], [10, 3, 3, 0]]])

data2 = np.array([[[11, 22, 33, 44], [22, 44, 55, 33], [0, -1, 22, 33]],
                  [[22, 44, 55, 11], [10, 55, 22, 22], [10, 33, 33, 0]]])

ht = {'CTYPE3': 'HPLT-TAN', 'CUNIT3': 'deg', 'CDELT3': 0.5, 'CRPIX3': 0, 'CRVAL3': 0, 'NAXIS3': 2,
      'CTYPE2': 'WAVE    ', 'CUNIT2': 'Angstrom', 'CDELT2': 0.2, 'CRPIX2': 0, 'CRVAL2': 0,
      'NAXIS2': 3,
      'CTYPE1': 'TIME    ', 'CUNIT1': 'min', 'CDELT1': 0.4, 'CRPIX1': 0, 'CRVAL1': 0, 'NAXIS1': 4}

hm = {
    'CTYPE1': 'WAVE    ', 'CUNIT1': 'Angstrom', 'CDELT1': 0.2, 'CRPIX1': 0, 'CRVAL1': 10,
    'NAXIS1': 4,
    'CTYPE2': 'HPLT-TAN', 'CUNIT2': 'deg', 'CDELT2': 0.5, 'CRPIX2': 2, 'CRVAL2': 0.5, 'NAXIS2': 3,
    'CTYPE3': 'HPLN-TAN', 'CUNIT3': 'deg', 'CDELT3': 0.4, 'CRPIX3': 2, 'CRVAL3': 1, 'NAXIS3': 2,
}

wt = WCS(header=ht, naxis=3)
wm = WCS(header=hm, naxis=3)

cube1 = NDCube(data, wt, missing_axis=[False, False, False, True], extra_coords=[
    ('pix', 0, u.Quantity(range(data.shape[0]), unit=u.pix)),
    ('distance', None, u.Quantity(0, unit=u.cm)),
    ('time', None, datetime.datetime(2000, 1, 1, 0, 0))])

cube2 = NDCube(data, wm, extra_coords=[
    ('pix', 0, u.Quantity(np.arange(1, data.shape[0]+1), unit=u.pix) +
     cube1.extra_coords['pix']['value'][-1]),
    ('distance', None, u.Quantity(1, unit=u.cm)),
    ('time', None, datetime.datetime(2000, 1, 1, 0, 1))])

cube3 = NDCube(data2, wt, missing_axis=[False, False, False, True], extra_coords=[
    ('pix', 0, u.Quantity(np.arange(1, data2.shape[0]+1), unit=u.pix) +
     cube2.extra_coords['pix']['value'][-1]),
    ('distance', None, u.Quantity(2, unit=u.cm)),
    ('time', None, datetime.datetime(2000, 1, 1, 0, 2))])

cube4 = NDCube(data2, wm, extra_coords=[
    ('pix', 0, u.Quantity(np.arange(1, data2.shape[0]+1), unit=u.pix) +
     cube3.extra_coords['pix']['value'][-1]),
    ('distance', None, u.Quantity(3, unit=u.cm)),
    ('time', None, datetime.datetime(2000, 1, 1, 0, 3))])

cube2_no_no = NDCube(data, wm, extra_coords=[
    ('pix', 0, u.Quantity(np.arange(1, data.shape[0]+1), unit=u.pix) +
     cube1.extra_coords['pix']['value'][-1]),
    ('time', None, datetime.datetime(2000, 1, 1, 0, 1))])

cube3_no_time = NDCube(data2, wt, missing_axis=[False, False, False, True], extra_coords=[
    ('pix', 0, u.Quantity(np.arange(1, data2.shape[0]+1), unit=u.pix) +
     cube2.extra_coords['pix']['value'][-1]),
    ('distance', None, u.Quantity(2, unit=u.cm))])

cube3_diff_compatible_unit = NDCube(
    data2, wt, missing_axis=[False, False, False, True], extra_coords=[
        ('pix', 0, u.Quantity(np.arange(data2.shape[0]), unit=u.pix) +
         cube2.extra_coords['pix']['value'][-1]),
        ('distance', None, u.Quantity(2, unit=u.cm).to('m')),
        ('time', None, datetime.datetime(2000, 1, 1, 0, 2))])

cube3_diff_incompatible_unit = NDCube(
    data2, wt, missing_axis=[False, False, False, True], extra_coords=[
        ('pix', 0, u.Quantity(np.arange(data2.shape[0]), unit=u.pix) +
         cube2.extra_coords['pix']['value'][-1]),
        ('distance', None, u.Quantity(2, unit=u.s)),
        ('time', None, datetime.datetime(2000, 1, 1, 0, 2))])

cube1_time_common = NDCube(
    data, wt, missing_axis=[False, False, False, True],
    extra_coords=[('time', 1, [datetime.datetime(2000, 1, 1) + datetime.timedelta(minutes=i)
                               for i in range(data.shape[1])])])

cube2_time_common = NDCube(data, wm, extra_coords=[
    ('time', 1,
     [cube1_time_common.extra_coords["time"]["value"][-1] + datetime.timedelta(minutes=i)
      for i in range(1, data.shape[1]+1)])])

cube1_no_extra_coords = NDCube(data, wt, missing_axis=[False, False, False, True])
cube3_no_extra_coords = NDCube(data2, wt, missing_axis=[False, False, False, True])

seq = NDCubeSequence([cube1, cube2, cube3, cube4], common_axis=0)
seq_bad_common_axis = NDCubeSequence([cube1, cube2, cube3, cube4], common_axis=None)
seq_time_common = NDCubeSequence([cube1_time_common, cube2_time_common], common_axis=1)
seq1 = NDCubeSequence([cube1, cube2, cube3, cube4])
seq2 = NDCubeSequence([cube1, cube2_no_no, cube3_no_time, cube4])
seq3 = NDCubeSequence([cube1, cube2, cube3_diff_compatible_unit, cube4])
seq4 = NDCubeSequence([cube1, cube2, cube3_diff_incompatible_unit, cube4])
seq_no_extra_coords = NDCubeSequence([cube1_no_extra_coords, cube3_no_extra_coords], common_axis=0)

nan_extra_coord = u.Quantity(range(4), unit=u.cm)
nan_extra_coord.value[1] = np.nan
nan_time_extra_coord = np.array([datetime.datetime(2000, 1, 1)+datetime.timedelta(minutes=i)
                                 for i in range(len(seq.data))])
nan_time_extra_coord[2] = np.nan


@pytest.mark.parametrize("test_input,expected", [
    (seq[0], NDCube),
    (seq[1], NDCube),
    (seq[2], NDCube),
    (seq[3], NDCube),
    (seq[0:1], NDCubeSequence),
    (seq[1:3], NDCubeSequence),
    (seq[0:2], NDCubeSequence),
    (seq[slice(0, 2)], NDCubeSequence),
    (seq[slice(0, 3)], NDCubeSequence),
])
def test_slice_first_index_sequence(test_input, expected):
    assert isinstance(test_input, expected)


@pytest.mark.parametrize("test_input,expected", [
    (seq[1:3], 2*u.pix),
    (seq[0:2], 2*u.pix),
    (seq[0::], 4*u.pix),
    (seq[slice(0, 2)], 2*u.pix),
    (seq[slice(0, 3)], 3*u.pix),
])
def test_slice_first_index_sequence(test_input, expected):
    assert test_input.dimensions[0] == expected


@pytest.mark.parametrize("test_input,expected", [
    (seq.index_as_cube[0:5].dimensions, (3*u.pix, [2., 2., 1.]*u.pix, 3*u.pix, 4*u.pix)),
    (seq.index_as_cube[1:3].dimensions, (2*u.pix, 1*u.pix, 3*u.pix, 4*u.pix)),
    (seq.index_as_cube[0:6].dimensions, (3*u.pix, 2*u.pix, 3*u.pix, 4*u.pix)),
    (seq.index_as_cube[0::].dimensions, (4*u.pix, 2*u.pix, 3*u.pix, 4*u.pix)),
    (seq.index_as_cube[0:5, 0].dimensions, (3*u.pix, [2., 2., 1.]*u.pix, 4*u.pix)),
    (seq.index_as_cube[1:3, 0:2].dimensions, (2*u.pix, 1*u.pix, 2*u.pix, 4*u.pix)),
    (seq.index_as_cube[0:6, 0, 0:1].dimensions, (3*u.pix, 2*u.pix, 1*u.pix)),
    (seq.index_as_cube[0::, 0, 0].dimensions, (4*u.pix, 2*u.pix)),
])
def test_index_as_cube(test_input, expected):
    for i in range(len(test_input)):
        try:
            assert test_input[i] == expected[i]
        except ValueError:
            assert (test_input[i].value == expected[i].value).all()
            assert test_input[i].unit == expected[i].unit


@pytest.mark.parametrize("test_input,expected", [
    (seq1.explode_along_axis(0), (8*u.pix, 3*u.pix, 4*u.pix)),
    (seq1.explode_along_axis(1), (12*u.pix, 2*u.pix, 4*u.pix)),
    (seq1.explode_along_axis(2), (16*u.pix, 2*u.pix, 3*u.pix)),
])
def test_explode_along_axis(test_input, expected):
    assert test_input.dimensions == expected


def test_explode_along_axis_error():
    with pytest.raises(ValueError):
        seq.explode_along_axis(1)


@pytest.mark.parametrize(
    "test_input,expected",
    [(seq, (4*u.pix, 2.*u.pix, 3.*u.pix, 4.*u.pix))])
def test_dimensions(test_input, expected):
    unit_tester = unittest.TestCase()
    unit_tester.assertEqual(test_input.dimensions, expected)


@pytest.mark.parametrize(
    "test_input,expected",
    [(seq, u.Quantity([8., 3., 4], unit=u.pix))])
def test_cube_like_dimensions(test_input, expected):
    assert (seq.cube_like_dimensions == expected).all()


@pytest.mark.parametrize("test_input", [(seq_bad_common_axis)])
def test_cube_like_dimensions_error(test_input):
    with pytest.raises(TypeError):
        seq_bad_common_axis.cube_like_dimensions


@pytest.mark.parametrize(
    "test_input,expected",
    [(seq, {'pix': u.Quantity([0., 1., 2., 3., 4., 5., 6., 7.], unit=u.pix)}),
     (seq_time_common,
      {'time': np.array([datetime.datetime(2000, 1, 1, 0, 0), datetime.datetime(2000, 1, 1, 0, 1),
                         datetime.datetime(2000, 1, 1, 0, 2), datetime.datetime(2000, 1, 1, 0, 3),
                         datetime.datetime(2000, 1, 1, 0, 4), datetime.datetime(2000, 1, 1, 0, 5)],
                        dtype=object)})])
def test_common_axis_extra_coords(test_input, expected):
    output = test_input.common_axis_extra_coords
    assert output.keys() == expected.keys()
    for key in output.keys():
        try:
            assert output[key] == expected[key]
        except ValueError:
            assert (output[key] == expected[key]).all()


@pytest.mark.parametrize("test_input", [(seq_no_extra_coords)])
def test_no_common_axis_extra_coords(test_input):
    assert seq_no_extra_coords.sequence_axis_extra_coords is None


@pytest.mark.parametrize(
    "test_input,expected",
    [(seq,
      {'distance': u.Quantity(range(4), unit=u.cm),
       'time': np.array([datetime.datetime(2000, 1, 1)+datetime.timedelta(minutes=i)
                        for i in range(len(seq.data))])}),
     (seq2,
      {'distance': nan_extra_coord,
       'time': nan_time_extra_coord}),
     (seq3,
      {'distance': u.Quantity(range(4), unit=u.cm),
       'time': np.array([datetime.datetime(2000, 1, 1)+datetime.timedelta(minutes=i)
                        for i in range(len(seq.data))])})])
def test_sequence_axis_extra_coords(test_input, expected):
    output = test_input.sequence_axis_extra_coords
    assert output.keys() == expected.keys()
    for key in output.keys():
        if isinstance(output[key], u.Quantity):
            assert output[key].unit == expected[key].unit
            np.testing.assert_array_almost_equal(output[key].value, expected[key].value)
        else:
            # For non-Quantities, must check element by element due to
            # possible mixture of NaN and non-number elements on
            # arrays, e.g. datetime does not work with
            # np.testing.assert_array_almost_equal().
            for i, output_value in enumerate(output[key]):
                if isinstance(output_value, float):
                    # Check if output is nan, expected is no and vice versa.
                    if not isinstance(expected[key][i], float):
                        raise AssertionError("{0} != {1}".format(output_value, expected[key][i]))
                    elif np.logical_xor(np.isnan(output_value), np.isnan(expected[key][i])):
                        raise AssertionError("{0} != {1}", format(output_value, expected[key][i]))
                    # Else assert they are equal if they are both not NaN.
                    elif not np.isnan(output_value):
                        assert output_value == expected[key][i]
                # Else, is output is not a float, assert it equals expected.
                else:
                    assert output_value == expected[key][i]


@pytest.mark.parametrize("test_input", [(seq_no_extra_coords)])
def test_no_sequence_axis_extra_coords(test_input):
    assert seq_no_extra_coords.sequence_axis_extra_coords is None


@pytest.mark.parametrize("test_input", [(seq4)])
def test_sequence_axis_extra_coords_incompatible_unit_error(test_input):
    with pytest.raises(u.UnitConversionError):
        test_input.sequence_axis_extra_coords



# -*- coding: utf-8 -*-
import pytest
import datetime
import copy

import numpy as np
import astropy.units as u
import matplotlib

from ndcube import NDCube, NDCubeSequence
from ndcube.utils.wcs import WCS
import ndcube.mixins.sequence_plotting

# sample data for tests
# TODO: use a fixture reading from a test file. file TBD.
data = np.array([[[1, 2, 3, 4], [2, 4, 5, 3], [0, -1, 2, 3]],
                 [[2, 4, 5, 1], [10, 5, 2, 2], [10, 3, 3, 0]]])

data2 = np.array([[[11, 22, 33, 44], [22, 44, 55, 33], [0, -1, 22, 33]],
                  [[22, 44, 55, 11], [10, 55, 22, 22], [10, 33, 33, 0]]])

ht = {'CTYPE3': 'HPLT-TAN', 'CUNIT3': 'deg', 'CDELT3': 0.5, 'CRPIX3': 0, 'CRVAL3': 0, 'NAXIS3': 2,
      'CTYPE2': 'WAVE    ', 'CUNIT2': 'Angstrom', 'CDELT2': 0.2, 'CRPIX2': 0, 'CRVAL2': 0,
      'NAXIS2': 3,
      'CTYPE1': 'TIME    ', 'CUNIT1': 'min', 'CDELT1': 0.4, 'CRPIX1': 0, 'CRVAL1': 0, 'NAXIS1': 4}
wt = WCS(header=ht, naxis=3)

hm = {
    'CTYPE1': 'WAVE    ', 'CUNIT1': 'Angstrom', 'CDELT1': 0.2, 'CRPIX1': 0, 'CRVAL1': 10,
    'NAXIS1': 4,
    'CTYPE2': 'HPLT-TAN', 'CUNIT2': 'deg', 'CDELT2': 0.5, 'CRPIX2': 2, 'CRVAL2': 0.5, 'NAXIS2': 3,
    'CTYPE3': 'HPLN-TAN', 'CUNIT3': 'deg', 'CDELT3': 0.4, 'CRPIX3': 2, 'CRVAL3': 1, 'NAXIS3': 2}
wm = WCS(header=hm, naxis=3)


cube1 = NDCube(
    data, wt, missing_axis=[False, False, False, True],
    extra_coords=[
        ('pix', 0, u.Quantity(range(data.shape[0]), unit=u.pix)),
        ('hi', 1, u.Quantity(range(data.shape[1]), unit=u.s)),
        ('distance', None, u.Quantity(0, unit=u.cm)),
        ('time', None, datetime.datetime(2000, 1, 1, 0, 0))])

cube1_with_unit = NDCube(
    data, wt, missing_axis=[False, False, False, True],
    unit=u.km,
    extra_coords=[
        ('pix', 0, u.Quantity(range(data.shape[0]), unit=u.pix)),
        ('hi', 1, u.Quantity(range(data.shape[1]), unit=u.s)),
        ('distance', None, u.Quantity(0, unit=u.cm)),
        ('time', None, datetime.datetime(2000, 1, 1, 0, 0))])

cube1_with_mask = NDCube(
    data, wt, missing_axis=[False, False, False, True],
    mask=np.zeros_like(data, dtype=bool),
    extra_coords=[
        ('pix', 0, u.Quantity(range(data.shape[0]), unit=u.pix)),
        ('hi', 1, u.Quantity(range(data.shape[1]), unit=u.s)),
        ('distance', None, u.Quantity(0, unit=u.cm)),
        ('time', None, datetime.datetime(2000, 1, 1, 0, 0))])

cube1_with_uncertainty = NDCube(
    data, wt, missing_axis=[False, False, False, True],
    uncertainty=np.sqrt(data),
    extra_coords=[
        ('pix', 0, u.Quantity(range(data.shape[0]), unit=u.pix)),
        ('hi', 1, u.Quantity(range(data.shape[1]), unit=u.s)),
        ('distance', None, u.Quantity(0, unit=u.cm)),
        ('time', None, datetime.datetime(2000, 1, 1, 0, 0))])

cube1_with_unit_and_uncertainty = NDCube(
    data, wt, missing_axis=[False, False, False, True],
    unit=u.km, uncertainty=np.sqrt(data),
    extra_coords=[
        ('pix', 0, u.Quantity(range(data.shape[0]), unit=u.pix)),
        ('hi', 1, u.Quantity(range(data.shape[1]), unit=u.s)),
        ('distance', None, u.Quantity(0, unit=u.cm)),
        ('time', None, datetime.datetime(2000, 1, 1, 0, 0))])

cube3 = NDCube(
    data2, wt, missing_axis=[False, False, False, True],
    extra_coords=[
        ('pix', 0, u.Quantity(np.arange(1, data2.shape[0]+1), unit=u.pix) +
         cube1.extra_coords['pix']['value'][-1]),
        ('hi', 1, u.Quantity(range(data2.shape[1]), unit=u.s)),
        ('distance', None, u.Quantity(2, unit=u.cm)),
        ('time', None, datetime.datetime(2000, 1, 1, 0, 2))])

cube3_with_unit = NDCube(
    data2, wt, missing_axis=[False, False, False, True],
    unit=u.m,
    extra_coords=[
        ('pix', 0, u.Quantity(np.arange(1, data2.shape[0]+1), unit=u.pix) +
         cube1.extra_coords['pix']['value'][-1]),
        ('hi', 1, u.Quantity(range(data2.shape[1]), unit=u.s)),
        ('distance', None, u.Quantity(2, unit=u.cm)),
        ('time', None, datetime.datetime(2000, 1, 1, 0, 2))])

cube3_with_mask = NDCube(
    data2, wt, missing_axis=[False, False, False, True],
    mask=np.zeros_like(data2, dtype=bool),
    extra_coords=[
        ('pix', 0, u.Quantity(np.arange(1, data2.shape[0]+1), unit=u.pix) +
         cube1.extra_coords['pix']['value'][-1]),
        ('hi', 1, u.Quantity(range(data2.shape[1]), unit=u.s)),
        ('distance', None, u.Quantity(2, unit=u.cm)),
        ('time', None, datetime.datetime(2000, 1, 1, 0, 2))])

cube3_with_uncertainty = NDCube(
    data2, wt, missing_axis=[False, False, False, True],
    uncertainty=np.sqrt(data2),
    extra_coords=[
        ('pix', 0, u.Quantity(np.arange(1, data2.shape[0]+1), unit=u.pix) +
         cube1.extra_coords['pix']['value'][-1]),
        ('hi', 1, u.Quantity(range(data2.shape[1]), unit=u.s)),
        ('distance', None, u.Quantity(2, unit=u.cm)),
        ('time', None, datetime.datetime(2000, 1, 1, 0, 2))])

cube3_with_unit_and_uncertainty = NDCube(
    data2, wt, missing_axis=[False, False, False, True],
    unit=u.m, uncertainty=np.sqrt(data2),
    extra_coords=[
        ('pix', 0, u.Quantity(np.arange(1, data2.shape[0]+1), unit=u.pix) +
         cube1.extra_coords['pix']['value'][-1]),
        ('hi', 1, u.Quantity(range(data2.shape[1]), unit=u.s)),
        ('distance', None, u.Quantity(2, unit=u.cm)),
        ('time', None, datetime.datetime(2000, 1, 1, 0, 2))])

cubem1 = NDCube(
    data, wm,
    extra_coords=[
        ('pix', 0, u.Quantity(range(data.shape[0]), unit=u.pix)),
        ('hi', 1, u.Quantity(range(data.shape[1]), unit=u.s)),
        ('distance', None, u.Quantity(0, unit=u.cm)),
        ('time', None, datetime.datetime(2000, 1, 1, 0, 0))])

cubem3 = NDCube(
    data2, wm,
    extra_coords=[
        ('pix', 0, u.Quantity(range(data.shape[0]), unit=u.pix)),
        ('hi', 1, u.Quantity(range(data.shape[1]), unit=u.s)),
        ('distance', None, u.Quantity(0, unit=u.cm)),
        ('time', None, datetime.datetime(2000, 1, 1, 0, 0))])

# Define some test NDCubeSequences.
common_axis = 0
seq = NDCubeSequence(data_list=[cube1, cube3, cube1, cube3], common_axis=common_axis)

seq_no_common_axis = NDCubeSequence(data_list=[cube1, cube3, cube1, cube3])

seq_with_units = NDCubeSequence(
    data_list=[cube1_with_unit, cube3_with_unit, cube1_with_unit, cube3_with_unit],
    common_axis=common_axis)

seq_with_masks = NDCubeSequence(
    data_list=[cube1_with_mask, cube3_with_mask, cube1_with_mask, cube3_with_mask],
    common_axis=common_axis)

seq_with_unit0 = NDCubeSequence(data_list=[cube1_with_unit, cube3,
                                           cube1_with_unit, cube3], common_axis=common_axis)

seq_with_mask0 = NDCubeSequence(data_list=[cube1_with_mask, cube3,
                                           cube1_with_mask, cube3], common_axis=common_axis)

seq_with_uncertainty = NDCubeSequence(data_list=[cube1_with_uncertainty, cube3_with_uncertainty,
                                                 cube1_with_uncertainty, cube3_with_uncertainty],
                                      common_axis=common_axis)

seq_with_some_uncertainty = NDCubeSequence(
    data_list=[cube1_with_uncertainty, cube3, cube1, cube3_with_uncertainty],
    common_axis=common_axis)

seq_with_units_and_uncertainty = NDCubeSequence(
    data_list=[cube1_with_unit_and_uncertainty, cube3_with_unit_and_uncertainty,
               cube1_with_unit_and_uncertainty, cube3_with_unit_and_uncertainty],
    common_axis=common_axis)

seq_with_units_and_some_uncertainty = NDCubeSequence(
    data_list=[cube1_with_unit_and_uncertainty, cube3_with_unit,
               cube1_with_unit, cube3_with_unit_and_uncertainty],
    common_axis=common_axis)

seq_with_some_masks = NDCubeSequence(data_list=[cube1_with_mask, cube3, cube1, cube3_with_mask],
                                     common_axis=common_axis)

seqm = NDCubeSequence(data_list=[cubem1, cubem3, cubem1, cubem3], common_axis=common_axis)

# Derive some expected data arrays in plot objects.
seq_data_stack = np.stack([cube.data for cube in seq_with_masks.data])
seq_mask_stack = np.stack([cube.mask for cube in seq_with_masks.data])

seq_stack = np.ma.masked_array(seq_data_stack, seq_mask_stack)
seq_stack_km = np.ma.masked_array(
    np.stack([(cube.data * cube.unit).to(u.km).value for cube in seq_with_units.data]),
    seq_mask_stack)

seq_data_concat = np.concatenate([cube.data for cube in seq_with_masks.data], axis=common_axis)
seq_mask_concat = np.concatenate([cube.mask for cube in seq_with_masks.data], axis=common_axis)

seq_concat = np.ma.masked_array(seq_data_concat, seq_mask_concat)
seq_concat_km = np.ma.masked_array(
    np.concatenate([(cube.data * cube.unit).to(u.km).value
                    for cube in seq_with_units.data], axis=common_axis),
    seq_mask_concat)

# Derive expected axis_ranges for non-cube-like cases.
x_axis_coords3 = np.array([0.4, 0.8, 1.2, 1.6]).reshape((1, 1, 4))
new_x_axis_coords3_shape = u.Quantity(seq.dimensions, unit=u.pix).value.astype(int)
new_x_axis_coords3_shape[-1] = 1
none_axis_ranges_axis3 = [np.arange(0, len(seq.data)+1),
                          np.array([0., 1., 2.]), np.arange(0, 4),
                          np.tile(np.array(x_axis_coords3), new_x_axis_coords3_shape)]
none_axis_ranges_axis0 = [np.arange(len(seq.data)),
                          np.array([0., 1., 2.]), np.arange(0, 4),
                          np.arange(0, int(seq.dimensions[-1].value)+1)]
distance0_none_axis_ranges_axis0 = [seq.sequence_axis_extra_coords["distance"].value,
                                    np.array([0., 1., 2.]), np.arange(0, 4),
                                    np.arange(0, int(seq.dimensions[-1].value)+1)]
distance0_none_axis_ranges_axis0_mm = [seq.sequence_axis_extra_coords["distance"].to("mm").value,
                                       np.array([0., 1., 2.]), np.arange(0, 4),
                                       np.arange(0, int(seq.dimensions[-1].value)+1)]
userrangequantity_none_axis_ranges_axis0 = [
    np.arange(int(seq.dimensions[0].value)), np.array([0., 1., 2.]), np.arange(0, 4),
    np.arange(0, int(seq.dimensions[-1].value)+1)]

userrangequantity_none_axis_ranges_axis0_1e7 = [
    (np.arange(int(seq.dimensions[0].value)) * u.J).to(u.erg).value, np.array([0., 1., 2.]),
    np.arange(0, 4), np.arange(0, int(seq.dimensions[-1].value)+1)]

hi2_none_axis_ranges_axis2 = [
    np.arange(0, len(seq.data)+1), np.array([0., 1., 2.]),
    np.arange(int(seq.dimensions[2].value)), np.arange(0, int(seq.dimensions[-1].value)+1)]

x_axis_coords1 = np.zeros(tuple([int(s.value) for s in seq.dimensions]))
x_axis_coords1[0, 1] = 1.
x_axis_coords1[1, 0] = 2.
x_axis_coords1[1, 1] = 3.
x_axis_coords1[2, 1] = 1.
x_axis_coords1[3, 0] = 2.
x_axis_coords1[3, 1] = 3.
pix1_none_axis_ranges_axis1 = [
    np.arange(0, len(seq.data)+1), x_axis_coords1, np.arange(0, 4),
    np.arange(0, int(seq.dimensions[-1].value)+1)]

# Derive expected extents
seq_axis1_lim_deg = [0.49998731, 0.99989848]
seq_axis1_lim_arcsec = [(axis1_xlim*u.deg).to(u.arcsec).value for axis1_xlim in seq_axis1_lim_deg]
seq_axis2_lim_m = [seq[:, :, :, 0].data[0].axis_world_coords()[-1][0].value,
                   seq[:, :, :, 0].data[0].axis_world_coords()[-1][-1].value]

# Derive expected axis_ranges for cube-like cases.
cube_like_new_x_axis_coords2_shape = u.Quantity(
    seq.cube_like_dimensions, unit=u.pix).value.astype(int)
cube_like_new_x_axis_coords2_shape[-1] = 1
cubelike_none_axis_ranges_axis2 = [
    np.arange(0, int(seq.cube_like_dimensions[0].value)+1), np.arange(0, 4),
    np.tile(x_axis_coords3, cube_like_new_x_axis_coords2_shape)]

cubelike_none_axis_ranges_axis2_s = copy.deepcopy(cubelike_none_axis_ranges_axis2)
cubelike_none_axis_ranges_axis2_s[2] = cubelike_none_axis_ranges_axis2_s[2] * 60.

cubelike_none_axis_ranges_axis0 = [[0, 8], np.arange(0, 4),
                                   np.arange(0, int(seq.cube_like_dimensions[-1].value)+1)]


@pytest.mark.parametrize("test_input, test_kwargs, expected_values", [
    (seq[:, 0, 0, 0], {},
     (np.arange(len(seq.data)), np.array([1, 11,  1, 11]),
      "meta.obs.sequence [None]", "Data [None]", (0, len(seq[:, 0, 0, 0].data)-1),
      (min([cube.data.min() for cube in seq[:, 0, 0, 0].data]),
       max([cube.data.max() for cube in seq[:, 0, 0, 0].data])))),

    (seq_with_units[:, 0, 0, 0], {},
     (np.arange(len(seq_with_units.data)), np.array([1, 0.011,  1, 0.011]),
      "meta.obs.sequence [None]", "Data [km]", (0, len(seq_with_units[:, 0, 0, 0].data)-1),
      (min([(cube.data * cube.unit).to(seq_with_units[:, 0, 0, 0].data[0].unit).value
            for cube in seq_with_units[:, 0, 0, 0].data]),
       max([(cube.data * cube.unit).to(seq_with_units[:, 0, 0, 0].data[0].unit).value
            for cube in seq_with_units[:, 0, 0, 0].data])))),

    (seq_with_uncertainty[:, 0, 0, 0], {},
     (np.arange(len(seq_with_uncertainty.data)), np.array([1, 11,  1, 11]),
      "meta.obs.sequence [None]", "Data [None]", (0, len(seq_with_uncertainty[:, 0, 0, 0].data)-1),
      (min([cube.data for cube in seq_with_uncertainty[:, 0, 0, 0].data]),
       max([cube.data for cube in seq_with_uncertainty[:, 0, 0, 0].data])))),

    (seq_with_units_and_uncertainty[:, 0, 0, 0], {},
     (np.arange(len(seq_with_units_and_uncertainty.data)), np.array([1, 0.011,  1, 0.011]),
      "meta.obs.sequence [None]", "Data [km]",
      (0, len(seq_with_units_and_uncertainty[:, 0, 0, 0].data)-1),
      (min([(cube.data*cube.unit).to(seq_with_units_and_uncertainty[:, 0, 0, 0].data[0].unit).value
            for cube in seq_with_units_and_uncertainty[:, 0, 0, 0].data]),
       max([(cube.data*cube.unit).to(seq_with_units_and_uncertainty[:, 0, 0, 0].data[0].unit).value
            for cube in seq_with_units_and_uncertainty[:, 0, 0, 0].data])))),

    (seq_with_units_and_some_uncertainty[:, 0, 0, 0], {},
     (np.arange(len(seq_with_units_and_some_uncertainty.data)), np.array([1, 0.011,  1, 0.011]),
      "meta.obs.sequence [None]", "Data [km]",
      (0, len(seq_with_units_and_some_uncertainty[:, 0, 0, 0].data)-1),
      (min([(cube.data*cube.unit).to(
          seq_with_units_and_some_uncertainty[:, 0, 0, 0].data[0].unit).value
          for cube in seq_with_units_and_some_uncertainty[:, 0, 0, 0].data]),
       max([(cube.data*cube.unit).to(
           seq_with_units_and_some_uncertainty[:, 0, 0, 0].data[0].unit).value
           for cube in seq_with_units_and_some_uncertainty[:, 0, 0, 0].data])))),

    (seq[:, 0, 0, 0], {"axes_coordinates": "distance"},
     ((seq.sequence_axis_extra_coords["distance"]), np.array([1, 11,  1, 11]),
      "distance [{0}]".format(seq.sequence_axis_extra_coords["distance"].unit), "Data [None]",
     (min(seq.sequence_axis_extra_coords["distance"].value),
      max(seq.sequence_axis_extra_coords["distance"].value)),
     (min([cube.data.min() for cube in seq[:, 0, 0, 0].data]),
      max([cube.data.max() for cube in seq[:, 0, 0, 0].data])))),

    (seq[:, 0, 0, 0], {"axes_coordinates": u.Quantity(np.arange(len(seq.data)), unit=u.cm),
                       "axes_units": u.km},
     (u.Quantity(np.arange(len(seq.data)), unit=u.cm).to(u.km), np.array([1, 11,  1, 11]),
      "meta.obs.sequence [km]", "Data [None]",
     (min((u.Quantity(np.arange(len(seq.data)), unit=u.cm).to(u.km).value)),
      max((u.Quantity(np.arange(len(seq.data)), unit=u.cm).to(u.km).value))),
     (min([cube.data.min() for cube in seq[:, 0, 0, 0].data]),
      max([cube.data.max() for cube in seq[:, 0, 0, 0].data]))))
    ])
def test_sequence_plot_1D_plot(test_input, test_kwargs, expected_values):
    # Unpack expected values
    expected_x_data, expected_y_data, expected_xlabel, expected_ylabel, \
      expected_xlim, expected_ylim = expected_values
    # Run plot method
    output = test_input.plot(**test_kwargs)
    # Check values are correct
    assert isinstance(output, matplotlib.axes.Axes)
    np.testing.assert_array_equal(output.lines[0].get_xdata(), expected_x_data)
    np.testing.assert_array_equal(output.lines[0].get_ydata(), expected_y_data)
    assert output.axes.get_xlabel() == expected_xlabel
    assert output.axes.get_ylabel() == expected_ylabel
    output_xlim = output.axes.get_xlim()
    assert output_xlim[0] <= expected_xlim[0]
    assert output_xlim[1] >= expected_xlim[1]
    output_ylim = output.axes.get_ylim()
    assert output_ylim[0] <= expected_ylim[0]
    assert output_ylim[1] >= expected_ylim[1]


@pytest.mark.parametrize("test_input, test_kwargs, expected_values", [
    (seq[:, :, 0, 0], {},
     (np.array([0.49998731, 0.99989848, 0.49998731, 0.99989848,
                0.49998731, 0.99989848, 0.49998731, 0.99989848]),
      np.array([1, 2, 11, 22,  1, 2, 11, 22]),
      "{0} [{1}]".format(seq[:, :, 0, 0].cube_like_world_axis_physical_types[common_axis], "deg"),
      "Data [None]", tuple(seq_axis1_lim_deg),
      (min([cube.data.min() for cube in seq[:, :, 0, 0].data]),
       max([cube.data.max() for cube in seq[:, :, 0, 0].data])))),

    (seq_with_units[:, :, 0, 0], {},
     (np.array([0.49998731, 0.99989848, 0.49998731, 0.99989848,
                0.49998731, 0.99989848, 0.49998731, 0.99989848]),
      np.array([1, 2, 0.011, 0.022,  1, 2, 0.011, 0.022]),
      "{0} [{1}]".format(seq[:, :, 0, 0].cube_like_world_axis_physical_types[common_axis], "deg"),
      "Data [km]", tuple(seq_axis1_lim_deg),
      (min([min((cube.data * cube.unit).to(u.km).value)
            for cube in seq_with_units[:, :, 0, 0].data]),
       max([max((cube.data * cube.unit).to(u.km).value)
            for cube in seq_with_units[:, :, 0, 0].data])))),

    (seq_with_uncertainty[:, :, 0, 0], {},
     (np.array([0.49998731, 0.99989848, 0.49998731, 0.99989848,
                0.49998731, 0.99989848, 0.49998731, 0.99989848]),
      np.array([1, 2, 11, 22,  1, 2, 11, 22]),
      "{0} [{1}]".format(
          seq_with_uncertainty[:, :, 0, 0].cube_like_world_axis_physical_types[
              common_axis], "deg"),
      "Data [None]", tuple(seq_axis1_lim_deg),
      (min([cube.data.min() for cube in seq_with_uncertainty[:, :, 0, 0].data]),
       max([cube.data.max() for cube in seq_with_uncertainty[:, :, 0, 0].data])))),

    (seq_with_some_uncertainty[:, :, 0, 0], {},
     (np.array([0.49998731, 0.99989848, 0.49998731, 0.99989848,
                0.49998731, 0.99989848, 0.49998731, 0.99989848]),
      np.array([1, 2, 11, 22,  1, 2, 11, 22]),
      "{0} [{1}]".format(
          seq_with_some_uncertainty[:, :, 0, 0].cube_like_world_axis_physical_types[
              common_axis], "deg"),
      "Data [None]", tuple(seq_axis1_lim_deg),
      (min([cube.data.min() for cube in seq_with_some_uncertainty[:, :, 0, 0].data]),
       max([cube.data.max() for cube in seq_with_some_uncertainty[:, :, 0, 0].data])))),

    (seq_with_units_and_uncertainty[:, :, 0, 0], {},
     (np.array([0.49998731, 0.99989848, 0.49998731, 0.99989848,
                0.49998731, 0.99989848, 0.49998731, 0.99989848]),
      np.array([1, 2, 0.011, 0.022,  1, 2, 0.011, 0.022]),
      "{0} [{1}]".format(
          seq_with_units_and_uncertainty[:, :, 0, 0].cube_like_world_axis_physical_types[
              common_axis], "deg"),
      "Data [km]", tuple(seq_axis1_lim_deg),
      (min([min((cube.data * cube.unit).to(u.km).value)
            for cube in seq_with_units[:, :, 0, 0].data]),
       max([max((cube.data * cube.unit).to(u.km).value)
            for cube in seq_with_units[:, :, 0, 0].data])))),

    (seq_with_units_and_some_uncertainty[:, :, 0, 0], {},
     (np.array([0.49998731, 0.99989848, 0.49998731, 0.99989848,
                0.49998731, 0.99989848, 0.49998731, 0.99989848]),
      np.array([1, 2, 0.011, 0.022,  1, 2, 0.011, 0.022]),
      "{0} [{1}]".format(
          seq_with_units_and_some_uncertainty[:, :, 0, 0].cube_like_world_axis_physical_types[
              common_axis], "deg"),
      "Data [km]", tuple(seq_axis1_lim_deg),
      (min([min((cube.data * cube.unit).to(u.km).value)
            for cube in seq_with_units[:, :, 0, 0].data]),
       max([max((cube.data * cube.unit).to(u.km).value)
            for cube in seq_with_units[:, :, 0, 0].data])))),

    (seq[:, :, 0, 0], {"axes_coordinates": "pix"},
     (seq[:, :, 0, 0].common_axis_extra_coords["pix"].value,
      np.array([1, 2, 11, 22,  1, 2, 11, 22]), "pix [pix]", "Data [None]",
      (min(seq[:, :, 0, 0].common_axis_extra_coords["pix"].value),
       max(seq[:, :, 0, 0].common_axis_extra_coords["pix"].value)),
      (min([cube.data.min() for cube in seq[:, :, 0, 0].data]),
       max([cube.data.max() for cube in seq[:, :, 0, 0].data])))),

    (seq[:, :, 0, 0],
     {"axes_coordinates": np.arange(10, 10+seq[:, :, 0, 0].cube_like_dimensions[0].value)},
     (np.arange(10, 10 + seq[:, :, 0, 0].cube_like_dimensions[0].value),
      np.array([1, 2, 11, 22,  1, 2, 11, 22]),
      "{0} [{1}]".format("", None), "Data [None]",
      (10, 10 + seq[:, :, 0, 0].cube_like_dimensions[0].value - 1),
      (min([cube.data.min() for cube in seq[:, :, 0, 0].data]),
       max([cube.data.max() for cube in seq[:, :, 0, 0].data]))))
    ])
def test_sequence_plot_as_cube_1D_plot(test_input, test_kwargs, expected_values):
    # Unpack expected values
    expected_x_data, expected_y_data, expected_xlabel, expected_ylabel, \
      expected_xlim, expected_ylim = expected_values
    # Run plot method
    output = test_input.plot_as_cube(**test_kwargs)
    # Check values are correct
    # Check type of ouput plot object
    assert isinstance(output, matplotlib.axes.Axes)
    # Check x and y data are correct.
    assert np.allclose(output.lines[0].get_xdata(), expected_x_data)
    assert np.allclose(output.lines[0].get_ydata(), expected_y_data)
    # Check x and y axis labels are correct.
    assert output.axes.get_xlabel() == expected_xlabel
    assert output.axes.get_ylabel() == expected_ylabel
    # Check all data is contained within x and y axes limits.
    output_xlim = output.axes.get_xlim()
    assert output_xlim[0] <= expected_xlim[0]
    assert output_xlim[1] >= expected_xlim[1]
    output_ylim = output.axes.get_ylim()
    assert output_ylim[0] <= expected_ylim[0]
    assert output_ylim[1] >= expected_ylim[1]


def test_sequence_plot_as_cube_error():
    with pytest.raises(TypeError):
        seq_no_common_axis.plot_as_cube()


@pytest.mark.parametrize("test_input, test_kwargs, expected_values", [
    (seq[:, :, 0, 0], {},
     (seq_stack[:, :, 0, 0],
      "custom:pos.helioprojective.lat [deg]", "meta.obs.sequence [None]",
      tuple(seq_axis1_lim_deg + [0, len(seq.data)-1]))),

    (seq_with_units[:, :, 0, 0], {},
     (seq_stack_km[:, :, 0, 0],
      "custom:pos.helioprojective.lat [deg]", "meta.obs.sequence [None]",
      tuple(seq_axis1_lim_deg + [0, len(seq.data)-1]))),

    (seq[:, :, 0, 0], {"plot_axis_indices": [0, 1]},
     (seq_stack[:, :, 0, 0].transpose(),
      "meta.obs.sequence [None]", "custom:pos.helioprojective.lat [deg]",
      tuple([0, len(seq.data)-1] + seq_axis1_lim_deg))),

    (seq[:, :, 0, 0], {"axes_coordinates": ["pix", "distance"]},
     (seq_stack[:, :, 0, 0],
      "pix [pix]", "distance [cm]",
      (min(seq[0, :, 0, 0].extra_coords["pix"]["value"].value),
       max(seq[0, :, 0, 0].extra_coords["pix"]["value"].value),
       min(seq[:, :, 0, 0].sequence_axis_extra_coords["distance"].value),
       max(seq[:, :, 0, 0].sequence_axis_extra_coords["distance"].value)))),
    # This example shows weakness of current extra coord axis values on 2D plotting!
    # Only the coordinates from the first cube are shown.

    (seq[:, :, 0, 0], {"axes_coordinates": [np.arange(
        10, 10+seq[:, :, 0, 0].dimensions[-1].value), "distance"], "axes_units": [None, u.m]},
     (seq_stack[:, :, 0, 0],
      " [None]", "distance [m]",
      (10, 10+seq[:, :, 0, 0].dimensions[-1].value-1,
       min(seq[:, :, 0, 0].sequence_axis_extra_coords["distance"].to(u.m).value),
       max(seq[:, :, 0, 0].sequence_axis_extra_coords["distance"].to(u.m).value)))),

    (seq[:, :, 0, 0], {"axes_coordinates": [np.arange(
        10, 10+seq[:, :, 0, 0].dimensions[-1].value)*u.deg, None], "axes_units": [u.arcsec, None]},
     (seq_stack[:, :, 0, 0],
      " [arcsec]", "meta.obs.sequence [None]",
      tuple(list(
          (np.arange(10, 10+seq[:, :, 0, 0].dimensions[-1].value)*u.deg).to(u.arcsec).value) \
          + [0, len(seq.data)-1])))
       ])
def test_sequence_plot_2D_image(test_input, test_kwargs, expected_values):
    # Unpack expected values
    expected_data, expected_xlabel, expected_ylabel, expected_extent = expected_values
    # Run plot method
    output = test_input.plot(**test_kwargs)
    # Check values are correct
    assert isinstance(output, matplotlib.axes.Axes)
    np.testing.assert_array_equal(output.images[0].get_array(), expected_data)
    assert output.xaxis.get_label_text() == expected_xlabel
    assert output.yaxis.get_label_text() == expected_ylabel
    assert np.allclose(output.images[0].get_extent(), expected_extent, rtol=1e-3)
    # Also check x and y values?????


@pytest.mark.parametrize("test_input, test_kwargs, expected_error", [
    (seq[:, :, 0, 0], {"axes_coordinates": [
        np.arange(10, 10+seq[:, :, 0, 0].dimensions[-1].value), None],
        "axes_units": [u.m, None]}, ValueError),

    (seq[:, :, 0, 0], {"axes_coordinates": [
        None, np.arange(10, 10+seq[:, :, 0, 0].dimensions[0].value)],
        "axes_units": [None, u.m]}, ValueError)
    ])
def test_sequence_plot_2D_image_errors(test_input, test_kwargs, expected_error):
    with pytest.raises(expected_error):
        output = test_input.plot(**test_kwargs)


@pytest.mark.parametrize("test_input, test_kwargs, expected_values", [
    (seq[:, :, :, 0], {},
     (seq_concat[:, :, 0],
      "em.wl [m]", "custom:pos.helioprojective.lat [deg]",
      tuple(seq_axis2_lim_m + seq_axis1_lim_deg))),

    (seq_with_units[:, :, :, 0], {},
     (seq_concat_km[:, :, 0],
      "em.wl [m]", "custom:pos.helioprojective.lat [deg]",
      tuple(seq_axis2_lim_m + seq_axis1_lim_deg))),

    (seq[:, :, :, 0], {"plot_axis_indices": [0, 1],
                       "axes_coordinates": ["pix", "hi"]},
     (seq_concat[:, :, 0].transpose(), "pix [pix]", "hi [s]",
      ((seq[:, :, :, 0].common_axis_extra_coords["pix"][0].value,
        seq[:, :, :, 0].common_axis_extra_coords["pix"][-1].value,
        seq[:, :, :, 0].data[0].extra_coords["hi"]["value"][0].value,
        seq[:, :, :, 0].data[0].extra_coords["hi"]["value"][-1].value)))),

    (seq[:, :, :, 0], {"axes_coordinates": [
        np.arange(10, 10+seq[:, :, :, 0].cube_like_dimensions[-1].value) * u.m,
        np.arange(10, 10+seq[:, :, :, 0].cube_like_dimensions[0].value) * u.m]},
     (seq_concat[:, :, 0], " [m]", " [m]",
      (10, 10+seq[:, :, :, 0].cube_like_dimensions[-1].value-1,
       10, 10+seq[:, :, :, 0].cube_like_dimensions[0].value-1))),

    (seq[:, :, :, 0], {"axes_coordinates": [
        np.arange(10, 10+seq[:, :, :, 0].cube_like_dimensions[-1].value) * u.m,
        np.arange(10, 10+seq[:, :, :, 0].cube_like_dimensions[0].value) * u.m],
        "axes_units": ["cm", u.cm]},
     (seq_concat[:, :, 0], " [cm]", " [cm]",
      (10*100, (10+seq[:, :, :, 0].cube_like_dimensions[-1].value-1)*100,
       10*100, (10+seq[:, :, :, 0].cube_like_dimensions[0].value-1)*100)))
    ])
def test_sequence_plot_as_cube_2D_image(test_input, test_kwargs, expected_values):
    # Unpack expected values
    expected_data, expected_xlabel, expected_ylabel, expected_extent = expected_values
    # Run plot method
    output = test_input.plot_as_cube(**test_kwargs)
    # Check values are correct
    assert isinstance(output, matplotlib.axes.Axes)
    np.testing.assert_array_equal(output.images[0].get_array(), expected_data)
    assert output.xaxis.get_label_text() == expected_xlabel
    assert output.yaxis.get_label_text() == expected_ylabel
    assert np.allclose(output.images[0].get_extent(), expected_extent, rtol=1e-3)
    # Also check x and y values?????


@pytest.mark.parametrize("test_input, test_kwargs, expected_error", [
    (seq[:, :, :, 0], {"axes_coordinates": [
        np.arange(10, 10+seq[:, :, :, 0].cube_like_dimensions[-1].value), None],
        "axes_units": [u.m, None]}, ValueError),

    (seq[:, :, :, 0], {"axes_coordinates": [
        None, np.arange(10, 10+seq[:, :, :, 0].cube_like_dimensions[0].value)],
        "axes_units": [None, u.m]}, ValueError)
    ])
def test_sequence_plot_as_cube_2D_image_errors(test_input, test_kwargs, expected_error):
    with pytest.raises(expected_error):
        output = test_input.plot_as_cube(**test_kwargs)


@pytest.mark.parametrize("test_input, test_kwargs, expected_data", [
    (seq, {}, seq_stack.reshape(4, 1, 2, 3, 4)),
    (seq_with_units, {}, seq_stack_km.reshape(4, 1, 2, 3, 4))
    ])
def test_sequence_plot_ImageAnimator(test_input, test_kwargs, expected_data):
    # Run plot method
    output = test_input.plot(**test_kwargs)
    # Check plot object properties are correct.
    assert isinstance(output, ndcube.mixins.sequence_plotting.ImageAnimatorNDCubeSequence)
    np.testing.assert_array_equal(output.data, expected_data)


@pytest.mark.parametrize("test_input, test_kwargs, expected_data", [
    (seq, {}, seq_concat.reshape(1, 8, 3, 4)),
    (seq_with_units, {}, seq_concat_km.reshape(1, 8, 3, 4))
    ])
def test_sequence_plot_as_cube_ImageAnimator(test_input, test_kwargs, expected_data):
    # Run plot method
    output = test_input.plot_as_cube(**test_kwargs)
    # Check plot object properties are correct.
    assert isinstance(output, ndcube.mixins.sequence_plotting.ImageAnimatorCubeLikeNDCubeSequence)
    np.testing.assert_array_equal(output.data, expected_data)


@pytest.mark.parametrize("test_input, expected", [
    ((seq_with_unit0.data, None), (None, None)),
    ((seq_with_unit0.data, u.km), (None, None)),
    ((seq_with_units.data, None), ([u.km, u.m, u.km, u.m], u.km)),
    ((seq_with_units.data, u.cm), ([u.km, u.m, u.km, u.m], u.cm))])
def test_determine_sequence_units(test_input, expected):
    output_seq_unit, output_unit = ndcube.mixins.sequence_plotting._determine_sequence_units(
        test_input[0], unit=test_input[1])
    assert output_seq_unit == expected[0]
    assert output_unit == expected[1]


def test_determine_sequence_units():
    with pytest.raises(ValueError):
        output_seq_unit, output_unit = ndcube.mixins.sequence_plotting._determine_sequence_units(
            seq.data, u.m)


@pytest.mark.parametrize("test_input, expected", [
    ((3, 1, "time", u.s), ([1], [None, 'time', None], [None, u.s, None])),
    ((3, None, None, None), ([-1, -2], None, None))])
def test_prep_axes_kwargs(test_input, expected):
    output = ndcube.mixins.sequence_plotting._prep_axes_kwargs(*test_input)
    for i in range(3):
        assert output[i] == expected[i]


@pytest.mark.parametrize("test_input, expected_error", [
    ((3, [0, 1, 2], ["time", "pix"], u.s), ValueError),
    ((3, 0, ["time", "pix"], u.s), ValueError),
    ((3, 0, "time", [u.s, u.pix]), ValueError),
    ((3, 0, 0, u.s), TypeError),
    ((3, 0, "time", 0), TypeError)])
def test_prep_axes_kwargs_errors(test_input, expected_error):
    with pytest.raises(expected_error):
        output = ndcube.mixins.sequence_plotting._prep_axes_kwargs(*test_input)


@pytest.mark.parametrize("test_input, test_kwargs, expected_values", [
    (seq, {"plot_axis_indices": 3},
     (seq_stack.data, none_axis_ranges_axis3, "time [min]", "Data [None]",
      (none_axis_ranges_axis3[-1].min(), none_axis_ranges_axis3[-1].max()),
      (seq_stack.data.min(), seq_stack.data.max()))),

    (seq_with_units, {"plot_axis_indices": -1, "data_unit": u.km},
     (seq_stack_km.data, none_axis_ranges_axis3, "time [min]", "Data [km]",
      (none_axis_ranges_axis3[-1].min(), none_axis_ranges_axis3[-1].max()),
      (seq_stack_km.data.min(), seq_stack_km.data.max()))),

    (seq_with_masks, {"plot_axis_indices": 0},
     (seq_stack, none_axis_ranges_axis0, "meta.obs.sequence [None]", "Data [None]",
      (none_axis_ranges_axis0[0].min(), none_axis_ranges_axis0[0].max()),
      (seq_stack.data.min(), seq_stack.data.max()))),

    (seq_with_some_masks, {"plot_axis_indices": 0},
     (seq_stack, none_axis_ranges_axis0, "meta.obs.sequence [None]", "Data [None]",
      (none_axis_ranges_axis0[0].min(), none_axis_ranges_axis0[0].max()),
      (seq_stack.data.min(), seq_stack.data.max()))),

    (seq, {"plot_axis_indices": 0, "axes_coordinates": "distance"},
     (seq_stack.data, distance0_none_axis_ranges_axis0, "distance [cm]", "Data [None]",
      (seq.sequence_axis_extra_coords["distance"].value.min(),
       seq.sequence_axis_extra_coords["distance"].value.max()),
      (seq_stack.data.min(), seq_stack.data.max()))),

    (seq, {"plot_axis_indices": 0, "axes_coordinates": "distance", "axes_units": "mm"},
     (seq_stack.data, distance0_none_axis_ranges_axis0_mm, "distance [mm]", "Data [None]",
      (seq.sequence_axis_extra_coords["distance"].to("mm").value.min(),
       seq.sequence_axis_extra_coords["distance"].to("mm").value.max()),
      (seq_stack.data.min(), seq_stack.data.max()))),

    (seq, {"plot_axis_indices": 0,
           "axes_coordinates": userrangequantity_none_axis_ranges_axis0[0]*u.J},
     (seq_stack.data, userrangequantity_none_axis_ranges_axis0, " [J]", "Data [None]",
      (userrangequantity_none_axis_ranges_axis0[0].min(),
       userrangequantity_none_axis_ranges_axis0[0].max()),
      (seq_stack.data.min(), seq_stack.data.max()))),

    (seq, {"plot_axis_indices": 0, "axes_units": u.erg,
           "axes_coordinates": userrangequantity_none_axis_ranges_axis0[0]*u.J},
     (seq_stack.data, userrangequantity_none_axis_ranges_axis0_1e7, " [erg]", "Data [None]",
      (userrangequantity_none_axis_ranges_axis0_1e7[0].min(),
       userrangequantity_none_axis_ranges_axis0_1e7[0].max()),
      (seq_stack.data.min(), seq_stack.data.max()))),

    (seq, {"plot_axis_indices": 2, "axes_coordinates": "hi"},
     (seq_stack.data, hi2_none_axis_ranges_axis2, "hi [s]", "Data [None]",
      (hi2_none_axis_ranges_axis2[2].min(), hi2_none_axis_ranges_axis2[2].max()),
      (seq_stack.data.min(), seq_stack.data.max()))),

    (seq, {"plot_axis_indices": 1, "axes_coordinates": "pix"},
     (seq_stack.data, pix1_none_axis_ranges_axis1, "pix [pix]", "Data [None]",
      (pix1_none_axis_ranges_axis1[1].min(), pix1_none_axis_ranges_axis1[1].max()),
      (seq_stack.data.min(), seq_stack.data.max())))
    ])
def test_sequence_plot_LineAnimator(test_input, test_kwargs, expected_values):
    # Unpack expected values
    expected_data, expected_axis_ranges, expected_xlabel, \
      expected_ylabel, expected_xlim, expected_ylim = expected_values
    # Run plot method.
    output = test_input.plot(**test_kwargs)
    # Check right type of plot object is produced.
    assert type(output) is ndcube.mixins.sequence_plotting.LineAnimatorNDCubeSequence
    # Check data being plotted is correct
    np.testing.assert_array_equal(output.data, expected_data)
    if type(expected_data) is np.ma.core.MaskedArray:
        np.testing.assert_array_equal(output.data.mask, expected_data.mask)
    # Check values of axes and sliders is correct.
    for i in range(len(output.axis_ranges)):
        print(i)
        assert np.allclose(output.axis_ranges[i], expected_axis_ranges[i])
    # Check plot axis labels and limits are correct
    assert output.xlabel == expected_xlabel
    assert output.ylabel == expected_ylabel
    assert output.xlim == expected_xlim
    assert output.ylim == expected_ylim


@pytest.mark.parametrize("test_input, test_kwargs, expected_values", [
    (seq, {"plot_axis_indices": 2, "axes_units": u.s},
     (seq_concat.data, cubelike_none_axis_ranges_axis2_s, "time [s]", "Data [None]",
      (cubelike_none_axis_ranges_axis2_s[2].min(), cubelike_none_axis_ranges_axis2_s[2].max()),
      (seq_concat.data.min(), seq_concat.data.max()))),

    (seq, {"plot_axis_indices": 0},
     (seq_concat.data, cubelike_none_axis_ranges_axis0,
      "custom:pos.helioprojective.lat [deg]", "Data [None]",
      (0, 7), (seq_concat.data.min(), seq_concat.data.max()))),

    (seq_with_masks, {"plot_axis_indices": 0},
     (seq_concat.data, cubelike_none_axis_ranges_axis0,
      "custom:pos.helioprojective.lat [deg]", "Data [None]",
      (0, 7), (seq_concat.data.min(), seq_concat.data.max()))),

    (seq_with_some_masks, {"plot_axis_indices": -3},
     (seq_concat.data, cubelike_none_axis_ranges_axis0,
      "custom:pos.helioprojective.lat [deg]", "Data [None]",
      (0, 7), (seq_concat.data.min(), seq_concat.data.max()))),

    (seqm, {"plot_axis_indices": 0},
     (seq_concat.data, cubelike_none_axis_ranges_axis0,
      "custom:pos.helioprojective.lon [deg]", "Data [None]",
      (0, 7), (seq_concat.data.min(), seq_concat.data.max())))
    ])
def test_sequence_plot_as_cube_LineAnimator(test_input, test_kwargs, expected_values):
    # Unpack expected values
    expected_data, expected_axis_ranges, expected_xlabel, \
      expected_ylabel, expected_xlim, expected_ylim = expected_values
    # Run plot method.
    output = test_input.plot_as_cube(**test_kwargs)
    # Check right type of plot object is produced.
    assert type(output) is ndcube.mixins.sequence_plotting.LineAnimatorCubeLikeNDCubeSequence
    # Check data being plotted is correct
    np.testing.assert_array_equal(output.data, expected_data)
    if type(expected_data) is np.ma.core.MaskedArray:
        np.testing.assert_array_equal(output.data.mask, expected_data.mask)
    # Check values of axes and sliders is correct.
    for i in range(len(output.axis_ranges)):
        assert np.allclose(output.axis_ranges[i], expected_axis_ranges[i])
    # Check plot axis labels and limits are correct
    assert output.xlabel == expected_xlabel
    assert output.ylabel == expected_ylabel
    assert output.xlim == expected_xlim
    assert output.ylim == expected_ylim



# -*- coding: utf-8 -*-

"""
Utilities for ndcube.
"""

import numpy as np

__all__ = ['wcs_axis_to_data_axis', 'data_axis_to_wcs_axis', 'select_order',
           'convert_extra_coords_dict_to_input_format', 'get_axis_number_from_axis_name']


def data_axis_to_wcs_axis(data_axis, missing_axis):
    """Converts a data axis number to the corresponding wcs axis number."""
    if data_axis is None:
        result = None
    else:
        if data_axis < 0:
            data_axis = np.invert(missing_axis).sum() + data_axis
        if data_axis > np.invert(missing_axis).sum()-1 or data_axis < 0:
            raise IndexError("Data axis out of range.  Number data axes = {0}".format(
                np.invert(missing_axis).sum()))
        result = len(missing_axis)-np.where(np.cumsum(
            [b is False for b in missing_axis][::-1]) == data_axis+1)[0][0]-1
    return result


def wcs_axis_to_data_axis(wcs_axis, missing_axis):
    """Converts a wcs axis number to the corresponding data axis number."""
    if wcs_axis is None:
        result = None
    else:
        if wcs_axis < 0:
            wcs_axis = len(missing_axis) + wcs_axis
        if wcs_axis > len(missing_axis)-1 or wcs_axis < 0:
            raise IndexError("WCS axis out of range.  Number WCS axes = {0}".format(
                len(missing_axis)))
        if missing_axis[wcs_axis]:
            result = None
        else:
            data_ordered_wcs_axis = len(missing_axis)-wcs_axis-1
            result = data_ordered_wcs_axis-sum(missing_axis[::-1][:data_ordered_wcs_axis])
    return result


def select_order(axtypes):
    """
    Returns indices of the correct data order axis priority given a list of WCS CTYPEs.

    For example, given ['HPLN-TAN', 'TIME', 'WAVE'] it will return
    [1, 2, 0] because index 1 (time) has the lowest priority, followed by
    wavelength and finally solar-x.

    Parameters
    ----------
    axtypes: str list
        The list of CTYPEs to be modified.

    """
    order = [(0, t) if t in ['TIME', 'UTC'] else
             (1, t) if t == 'WAVE' else
             (2, t) if t == 'HPLT-TAN' else
             (axtypes.index(t) + 3, t) for t in axtypes]
    order.sort()
    result = [axtypes.index(s) for (_, s) in order]
    return result


def _format_input_extra_coords_to_extra_coords_wcs_axis(extra_coords, missing_axis,
                                                        data_shape):
    extra_coords_wcs_axis = {}
    coord_format_error = ("Coord must have three properties supplied, "
                          "name (str), axis (int), values (Quantity or array-like)."
                          " Input coord: {0}")
    coord_0_format_error = ("1st element of extra coordinate tuple must be a "
                            "string giving the coordinate's name.")
    coord_1_format_error = ("2nd element of extra coordinate tuple must be None "
                            "or an int giving the data axis "
                            "to which the coordinate corresponds.")
    coord_len_error = ("extra coord ({0}) must have same length as data axis "
                       "to which it is assigned: coord length, {1} != data axis length, {2}")
    for coord in extra_coords:
        # Check extra coord has the right number and types of info.
        if len(coord) != 3:
            raise ValueError(coord_format_error.format(coord))
        if not isinstance(coord[0], str):
            raise ValueError(coord_0_format_error.format(coord))
        if coord[1] is not None and not isinstance(coord[1], int) and \
                not isinstance(coord[1], np.int64):
            raise ValueError(coord_1_format_error)
        # Unless extra coord corresponds to a missing axis, check length
        # of coord is same is data axis to which is corresponds.
        if coord[1] is not None:
            if not missing_axis[::-1][coord[1]]:

                if len(coord[2]) != data_shape[coord[1]]:
                    raise ValueError(coord_len_error.format(coord[0], len(coord[2]),
                                                            data_shape[coord[1]]))
        # Determine wcs axis corresponding to data axis of coord
        extra_coords_wcs_axis[coord[0]] = {
            "wcs axis": data_axis_to_wcs_axis(coord[1], missing_axis),
            "value": coord[2]}
    return extra_coords_wcs_axis


def convert_extra_coords_dict_to_input_format(extra_coords, missing_axis):
        """
        Converts NDCube.extra_coords attribute to format required as input for new NDCube.

        Parameters
        ----------
        extra_coords: dict
            An NDCube.extra_coords instance.

        Returns
        -------
        input_format: `list`
            Infomation on extra coords in format required by `NDCube.__init__`.

        """
        coord_names = list(extra_coords.keys())
        result = []
        for name in coord_names:
            coord_keys = list(extra_coords[name].keys())
            if "wcs axis" in coord_keys and "axis" not in coord_keys:
                axis = wcs_axis_to_data_axis(extra_coords[name]["wcs axis"], missing_axis)
            elif "axis" in coord_keys and "wcs axis" not in coord_keys:
                axis = extra_coords[name]["axis"]
            else:
                raise KeyError("extra coords dict can have keys 'wcs axis' or 'axis'.  Not both.")
            result.append((name, axis, extra_coords[name]["value"]))
        return result


def get_axis_number_from_axis_name(axis_name, world_axis_physical_types):
    """
    Returns axis number (numpy ordering) given a substring unique to a world axis type string.

    Parameters
    ----------
    axis_name: `str`
        Name or substring of name of axis as defined by NDCube.world_axis_physical_types

    world_axis_physical_types: iterable of `str`
        Output from NDCube.world_axis_physical_types for relevant cube,
        i.e. iterable of string axis names.

    Returns
    -------
    axis_index[0]: `int`
        Axis number (numpy ordering) corresponding to axis name
    """
    axis_index = [axis_name in world_axis_type for world_axis_type in world_axis_physical_types]
    axis_index = np.arange(len(world_axis_physical_types))[axis_index]
    if len(axis_index) != 1:
        raise ValueError("User defined axis with a string that is not unique to "
                         "a physical axis type. {0} not in any of {1}".format(
                             axis_name, world_axis_physical_types))
    return axis_index[0]



# Autogenerated by Astropy-affiliated package astropy_helpers's setup.py on 2017-08-16 14:10:37.888812
from __future__ import unicode_literals
import datetime

version = "2.0.1"
githash = "14ca346b0da3e92e65bdc398cc8f726cbd7be57d"


major = 2
minor = 0
bugfix = 1

release = True
timestamp = datetime.datetime(2017, 8, 16, 14, 10, 37, 888812)
debug = False

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"



# Licensed under a 2-clause BSD style license - see LICENSE.rst

"""
This is a SunPy affiliated package.
"""

# Affiliated packages may add whatever they like to this file, but
# should keep this content at the top.
# ----------------------------------------------------------------------------
from ._sunpy_init import *
# ----------------------------------------------------------------------------

if not _ASTROPY_SETUP_:
    # For egg_info test builds to pass, put package imports here.
    from .ndcube import NDCube, NDCubeOrdered
    from .ndcube_sequence import NDCubeSequence



# -*- coding: utf-8 -*-

"""
Utilities for ndcube sequence.
"""

from copy import deepcopy
from collections import namedtuple
from functools import singledispatch

import numpy as np
import astropy.units as u


__all__ = ['SequenceSlice', 'SequenceItem', 'slice_sequence', 'convert_item_to_sequence_items',
           'convert_cube_like_item_to_sequence_items', 'convert_slice_nones_to_ints']


SequenceSlice = namedtuple("SequenceSlice", "sequence_index common_axis_item")
"""
Define SequenceSlice named tuple of length 2. Its attributes are:
sequence_index: an int giving the index of a cube within an NDCubeSequence.
common_axis_item: slice of int index of to be to be applied to the common
axis of the cube.
"""
SequenceItem = namedtuple("SequenceItem", "sequence_index cube_item")
"""
Define SequenceItem named tuple of length 2. Its attributes are:
sequence_index: an int giving the index of a cube within an NDCubeSequence.
cube_item: item (int, slice, tuple) to be applied to cube identified
by sequence_index attribute.
"""


def slice_sequence(cubesequence, item):
    """
    Slice an NDCubeSequence given a slicing/index item.

    Parameters
    ----------
    cubesequence: `ndcube.NDCubeSequence`
        The cubesequence to slice.

    item: `int`, `slice`, or `tuple` of `int` and/or `slice`.
        An slice/index item compatible with input to NDCubeSequence.__getitem__.

    Returns
    -------
    result: `NDCubeSequence` or `NDCube`
        The sliced cube sequence.

    """
    if item is None or (isinstance(item, tuple) and None in item):
        raise IndexError("None indices not supported")
    # Convert item to list of SequenceSlices
    sequence_items = convert_item_to_sequence_items(item, len(cubesequence.data))
    return slice_sequence_by_sequence_items(cubesequence, sequence_items)


@singledispatch
def convert_item_to_sequence_items(item, n_cubes=None, cube_item=None):
    """
    Converts NDCubeSequence __getitem__ item to list of SequenceSlice objects.

    Parameters
    ----------
    item: `int`, `slice`, or `tuple` of `int` and/or `slice`.
        An slice/index item compatible with input to NDCubeSequence.__getitem__.

    n_cubes: `int`
        Number of cubes in NDCubeSequence being sliced.  Must be supplied, but
        not used if item type is `int` or `slice`.

    Returns
    -------
    result: `list` of SequenceItem `namedtuple`.
        The slice/index items for each relevant NDCube within the NDCubeSequence
        which together represent the original input slice/index item.

    """
    # If type if the first input of this function does not match the
    # type of first input of one of the below registered functions,
    # raise an error.  Otherwise one of the below registered functions
    # is executed.
    raise TypeError("Unrecognized slice type: {0}", item)


@convert_item_to_sequence_items.register(int)
def _get_sequence_items_from_int_item(int_item, n_cubes=None, cube_item=slice(None)):
    """
    Converts int index of an NDCubeSequence to list of SequenceSlices.

    Parameters
    ----------
    int_item: `int`
        index of NDCube within NDCubeSequence to be slices out.

    n_cubes: `None`
        Not used.  Exists in API to be consistent with API of convert_item_to_sequence_items()
        to which it this function is registered under single dispatch.

    cube_item: `int`, `slice`, or `tuple`
        Item to be applied to selected NDCube.

    Returns
    -------
    result: `list` of SequenceItem `namedtuple`
        The slice/index items for each relevant NDCube within the NDCubeSequence
        which together represent the original input slice/index item.

    """
    return [SequenceItem(int_item, cube_item)]


@convert_item_to_sequence_items.register(slice)
def _get_sequence_items_from_slice_item(slice_item, n_cubes, cube_item=slice(None)):
    """
    Converts slice item of an NDCubeSequence to list of SequenceSlices.

    Parameters
    ----------
    slice_item: `slice`
        Indicates which NDCubes within NDCubeSequence are to be slices out.

    n_cubes: `int`
        Number of cubes in NDCubeSequence being sliced.

    cube_item: `int`, `slice`, or `tuple`
        Item to be applied to each selected NDCube.

    Returns
    -------
    sequence_items: `list` of SequenceItem `namedtuple`.
        The slice/index items for each relevant NDCube within the NDCubeSequence
        which together represent the original input slice/index item.

    """
    # If there are None types in slice, replace with correct entries based on sign of step.
    no_none_slice = convert_slice_nones_to_ints(slice_item, n_cubes)
    # Derive SequenceItems for each cube.  Recall that
    # once convert_slice_nones_to_ints() has been applied, a None will
    # only be present to signify the beginning of the array when the
    # step is negative.  Therefore, if the stop parmeter of the above
    # slice object is None, set the stop condition of the below for
    # loop to -1.
    if no_none_slice.stop is None:
        stop = -1
    else:
        stop = no_none_slice.stop
    sequence_items = [SequenceItem(i, cube_item)
                      for i in range(no_none_slice.start, stop, no_none_slice.step)]
    return sequence_items


@convert_item_to_sequence_items.register(tuple)
def _get_sequence_items_from_tuple_item(tuple_item, n_cubes, cube_item=None):
    """
    Converts NDCubeSequence slice item tuple to list of SequenceSlice objects.

    Parameters
    ----------
    tuple_item: `tuple` of `int` and/or `slice`.
        Index/slice for different dimensions of NDCubeSequence.  The first entry
        applies to the sequence axis while subsequent entries make up the slicing
        item to be applied to the NDCubes.

    n_cubes: `int`
        Number of cubes in NDCubeSequence being sliced.

    cube_item: `None`
        Not used.  Exists in API to be consistent with API of convert_item_to_sequence_items()
        to which it this function is registered under single dispatch.

    Returns
    -------
    sequence_items: `list` of SequenceItem `namedtuple`.
        The slice/index items for each relevant NDCube within the NDCubeSequence
        which together represent the original input slice/index item.

    """
    # Define slice to be applied to cubes.
    if len(tuple_item[1:]) == 1:
        cube_item = tuple_item[1]
    else:
        cube_item = tuple_item[1:]
    # Based on type of sequence index, define sequence slices.
    sequence_items = convert_item_to_sequence_items(tuple_item[0], n_cubes=n_cubes,
                                                    cube_item=cube_item)
    return sequence_items


def slice_sequence_by_sequence_items(cubesequence, sequence_items):
    """
    Slices an NDCubeSequence given a list of SequenceSlice objects.

    Parameters
    ----------
    cubesequence: `ndcube.NDCubeSequence`
        The cubesequence to slice.
    sequence_items: `list` of `SequenceItem`
        Slices to be applied to each relevant NDCube in the sequence.

    Returns
    -------
    result: `NDCubeSequence` or `NDCube`
        The sliced cube sequence.

    """
    result = deepcopy(cubesequence)
    if len(sequence_items) == 1:
        return result.data[sequence_items[0].sequence_index][sequence_items[0].cube_item]
    else:
        data = [result.data[sequence_item.sequence_index][sequence_item.cube_item]
                for sequence_item in sequence_items]
        result.data = data
        return result


def _index_sequence_as_cube(cubesequence, item):
    """
    Enables NDCubeSequence to be indexed as if it were a single NDCube.

    This is only possible if cubesequence._common_axis is set,
    i.e. if the cubes are sequenced in order along one of the cube axes.
    For example, if cubesequence._common_axis is 1 where the first axis is
    time, and the cubes are sequenced chronologically such that the last
    time slice of one cube is directly followed in time by the first time
    slice of the next cube, then this function allows the NDCubeSequence to
    be indexed as though all cubes were combined into one ordered along
    the time axis.

    Parameters
    ----------
    cubesequence: `ndcube.NDCubeSequence`
        The cubesequence to get the item from
    item: `int`, `slice` or `tuple` of `int` and/or `slice`.
        The item to get from the cube.  If tuple length must be <= number
        of dimensions in single cube.

    Example
    -------
    >>> # Say we have three Cubes each cube has common_axis=1 is time and shape=(3,3,3)
    >>> data_list = [cubeA, cubeB, cubeC] # doctest: +SKIP
    >>> cs = NDCubeSequence(data_list, meta=None, common_axis=1) # doctest: +SKIP
    >>> # return zeroth time slice of cubeB in via normal CubeSequence indexing.
    >>> cs[1,:,0,:] # doctest: +SKIP
    >>> # Return same slice using this function
    >>> _index_sequence_as_cube(cs, (slice(0, cubeB.shape[0]), 0,
    ...                             (slice(0, cubeB.shape[2])) # doctest: +SKIP

    """
    # Convert index_as_cube item to a list of regular NDCubeSequence
    # items of each relevant cube.
    common_axis_cube_lengths = np.array([c.data.shape[cubesequence._common_axis]
                                         for c in cubesequence.data])
    sequence_items = convert_cube_like_item_to_sequence_items(item, cubesequence._common_axis,
                                                              common_axis_cube_lengths)
    # Use sequence items to slice NDCubeSequence.
    return slice_sequence_by_sequence_items(cubesequence, sequence_items)


def convert_cube_like_item_to_sequence_items(cube_like_item, common_axis, common_axis_cube_lengths):
    """
    Converts an input item to NDCubeSequence.index_as_cube to a list od SequenceSlice objects.

    Parameters
    ----------
    cube_like_item: `int`, `slice`, of `tuple` of `int and/or `slice`.
        Item compatible with input to NDCubeSequence.index_as_cube.

    common_axis: `int`
        Data axis of NDCubes common to NDCubeSequence

    common_axis_cube_lengths: `np.array`
        Length of each cube in sequence along the common axis.

    Returns
    -------
    sequence_items: `list` of SequenceItem `namedtuple`.
        The slice/index items for each relevant NDCube within the NDCubeSequence
        which together represent the original input slice/index item.

    """
    invalid_item_error_message = "Invalid index/slice input."
    # Case 1: Item is int and common axis is 0.
    if isinstance(cube_like_item, int):
        if common_axis != 0:
            raise ValueError("Input can only be indexed with an int if "
                             "CubeSequence's common axis is 0. common "
                             "axis = {0}".format(common_axis))
        else:
            # Derive list of SequenceSlice objects that describes the
            # cube_like_item in regular slicing notation.
            sequence_slices = [_convert_cube_like_index_to_sequence_slice(
                cube_like_item, common_axis_cube_lengths)]
            all_axes_item = None
    # Case 2: Item is slice and common axis is 0.
    elif isinstance(cube_like_item, slice):
        if common_axis != 0:
            raise ValueError("Input can only be sliced with a single slice if "
                             "CubeSequence's common axis is 0. common "
                             "axis = {0}".format(common_axis))
        else:
            # Derive list of SequenceSlice objects that describes the
            # cube_like_item in regular slicing notation.
            # First ensure None types within slice are replaced with appropriate ints.
            sequence_slices = _convert_cube_like_slice_to_sequence_slices(
                cube_like_item, common_axis_cube_lengths)
            all_axes_item = None
    # Case 3: Item is tuple.
    elif isinstance(cube_like_item, tuple):
        # Check item is long enough to include common axis.
        if len(cube_like_item) < common_axis+1:
            raise ValueError("Input item not long enough to include common axis."
                             "Must have length > {0}".format(common_axis))
        # Based on type of slice/index in the common axis position of
        # the cube_like_item, derive list of SequenceSlice objects that
        # describes the cube_like_item in regular slicing notation.
        if isinstance(cube_like_item[common_axis], int):
            sequence_slices = [_convert_cube_like_index_to_sequence_slice(
                cube_like_item[common_axis], common_axis_cube_lengths)]
        elif isinstance(cube_like_item[common_axis], slice):
            sequence_slices = _convert_cube_like_slice_to_sequence_slices(
                cube_like_item[common_axis], common_axis_cube_lengths)
        else:
            raise ValueError(invalid_item_error_message)
        all_axes_item = cube_like_item
    else:
        raise TypeError("Unrecognized item type.")
    # Convert the sequence slices, that only describe the slicing along
    # the sequence axis and common axis to sequence items which
    # additionally describe how the non-common cube axes should be sliced.
    sequence_items = [_convert_sequence_slice_to_sequence_item(
        sequence_slice, common_axis, cube_like_item=all_axes_item)
        for sequence_slice in sequence_slices]
    return sequence_items


def _convert_cube_like_index_to_sequence_slice(cube_like_index, common_axis_cube_lengths):
    """
    Converts a cube-like index of an NDCubeSequence to indices along the sequence and common axes.

    Parameters
    ----------
    cube_like_index: `int`
        Cube-like index of NDCubeSequence

    common_axis_cube_lengths: iterable of `int`
        Length of each cube along common axis.

    Returns
    -------
    sequence_slice: SequenceSlice `namedtuple`.
        First element gives index of cube along sequence axis.
        Second element each index along common axis of relevant cube.

    """
    # Derive cumulative lengths of cubes along common axis.
    cumul_common_axis_cube_lengths = np.cumsum(common_axis_cube_lengths)
    # If cube_like_index is within 0th cube in sequence, it is
    # simple to determine the sequence and common axis indices.
    try:
        index_in_0th_cube = cube_like_index < cumul_common_axis_cube_lengths[0]
    except TypeError as err:
        none_not_int_error_messages = [
            "'>' not supported between instances of 'int' and 'NoneType'",
            "unorderable types: int() > NoneType()"]
        if err.args[0] in none_not_int_error_messages:
            index_in_0th_cube = True
        else:
            raise err
    if index_in_0th_cube:
        sequence_index = 0
        common_axis_index = cube_like_index
    # Else use more in-depth method.
    else:
        # Determine the index of the relevant cube within the sequence
        # from the cumulative common axis cube lengths.
        sequence_index = np.where(cumul_common_axis_cube_lengths <= cube_like_index)[0][-1]
        if cube_like_index > cumul_common_axis_cube_lengths[-1]-1:
            # If the cube is out of range then return the last common axis index.
            common_axis_index = common_axis_cube_lengths[-1]
        else:
            # Else use simple equation to derive the relevant common axis index.
            common_axis_index = cube_like_index - cumul_common_axis_cube_lengths[sequence_index]
        # sequence_index should be plus one as the sequence_index earlier is
        # previous index if it is not already the last cube index.
        if sequence_index < cumul_common_axis_cube_lengths.size - 1:
            sequence_index += 1
    # Return sequence and cube indices.  Ensure they are int, rather
    # than np.int64 to avoid confusion in checking type elsewhere.
    if common_axis_index is not None:
        common_axis_index = int(common_axis_index)
    return SequenceSlice(int(sequence_index), common_axis_index)


def _convert_cube_like_slice_to_sequence_slices(cube_like_slice, common_axis_cube_lengths):
    """
    Converts common axis slice input to NDCubeSequence.index_as_cube to a list of sequence indices.

    Parameters
    ----------
    cube_like_slice: `slice`
        Slice along common axis in NDCubeSequence.index_as_cube item.

    common_axis_cube_lengths: iterable of `int`
        Length of each cube along common axis.

    Returns
    -------
    sequence_slices: `list` of SequenceSlice `namedtuple`.
        List sequence slices (sequence axis, common axis) for each element
        along common axis represented by input cube_like_slice.

    """
    # Ensure any None attributes in input slice are filled with appropriate ints.
    cumul_common_axis_cube_lengths = np.cumsum(common_axis_cube_lengths)
    # Determine sequence indices of cubes included in cube-like slice.
    cube_like_indices = np.arange(cumul_common_axis_cube_lengths[-1])[cube_like_slice]
    n_cube_like_indices = len(cube_like_indices)
    one_step_sequence_slices = np.empty(n_cube_like_indices, dtype=object)
    # Define array of ints for all indices along common axis.
    # This is restricted to range of interest below.
    sequence_int_indices = np.zeros(n_cube_like_indices, dtype=int)
    for i in range(n_cube_like_indices):
        one_step_sequence_slices[i] = _convert_cube_like_index_to_sequence_slice(
            cube_like_indices[i], common_axis_cube_lengths)
        sequence_int_indices[i] = one_step_sequence_slices[i].sequence_index
    unique_index = np.sort(np.unique(sequence_int_indices, return_index=True)[1])
    unique_sequence_indices = sequence_int_indices[unique_index]
    # Convert start and stop cube-like indices to sequence indices.
    first_sequence_index = _convert_cube_like_index_to_sequence_slice(cube_like_slice.start,
                                                                      common_axis_cube_lengths)
    last_sequence_index = _convert_cube_like_index_to_sequence_slice(cube_like_slice.stop,
                                                                     common_axis_cube_lengths)
    # Since the last index of any slice represents
    # 'up to but not including this element', if the last sequence index
    # is the first element of a new cube, elements from the last cube
    # will not appear in the sliced sequence.  Therefore for ease of
    # slicing, we can redefine the final sequence index as the penultimate
    # cube and its common axis index as beyond the range of the
    # penultimate cube's length along the common axis.
    if (last_sequence_index.sequence_index > first_sequence_index.sequence_index and
            last_sequence_index.common_axis_item == 0):
        last_sequence_index = SequenceSlice(
            last_sequence_index.sequence_index-1,
            common_axis_cube_lengths[last_sequence_index.sequence_index-1])
    # Iterate through relevant cubes and determine slices for each.
    # Do last cube outside loop as its end index may not correspond to
    # the end of the cube's common axis.
    if cube_like_slice.step is None:
        step = 1
    else:
        step = cube_like_slice.step
    sequence_slices = []
    common_axis_start_index = first_sequence_index.common_axis_item
    j = 0
    while j < len(unique_sequence_indices)-1:
        # Let i be the index along the sequence axis of the next relevant cube.
        i = unique_sequence_indices[j]
        # Determine last common axis index for this cube.
        common_axis_last_index = common_axis_cube_lengths[i] - (
            (common_axis_cube_lengths[i] - common_axis_start_index) % step)
        # Generate SequenceSlice for this cube and append to list.
        sequence_slices.append(SequenceSlice(
            i, slice(common_axis_start_index,
                     min(common_axis_last_index+1, common_axis_cube_lengths[i]), step)))
        # Determine first common axis index for next cube.
        if common_axis_cube_lengths[i] == common_axis_last_index:
            common_axis_start_index = step-1
        else:
            common_axis_start_index = \
              step - (((common_axis_cube_lengths[i] - common_axis_last_index) % step) +
                      cumul_common_axis_cube_lengths[unique_sequence_indices[j+1]-1] -
                      cumul_common_axis_cube_lengths[i])
        # Iterate counter.
        j += 1
    # Create slice for last cube manually.
    sequence_slices.append(SequenceSlice(
        unique_sequence_indices[j],
        slice(common_axis_start_index, last_sequence_index.common_axis_item, step)))
    return sequence_slices


def _convert_sequence_slice_to_sequence_item(sequence_slice, common_axis, cube_like_item=None):
    """
    Converts sequence/cube index to a SequenceSlice object.

    Parameters
    ----------
    sequence_slice: SequenceSlice `namedtuple`
        0th element gives index of cube along sequence axis.
        1st element each index along common axis of relevant cube.
        Must be same format as output from _convert_cube_like_index_to_sequence_slice.

    common_axis: `int`
        Common axis as defined in NDCubeSequence.

    cube_like_item: `None` or `tuple` of `slice` and/or `int` objects (Optional)
        The original item input to `NDCubeSequence.index_as_cube` including the
        slices/indices of non-common axes of cubes within sequence.  If None, a
        tuple of slice(None) objects is generated  long enough so that the last
        element in the tuple corresponds to the common axis and is set to the
        1st (0-based counting) the sequence_index input, above.  This tuple is
        then set to the cube_item attribute of the output `SequenceSlice` object.

    Returns
    -------
    sequence_item: SequenceItem `namedtuple`.
        Describes sequence index of an NDCube within an NDCubeSequence and the
        slice/index item to be applied to the whole NDCube.

    """
    if cube_like_item is None and common_axis == 0:
        sequence_item = SequenceItem(sequence_slice.sequence_index,
                                     sequence_slice.common_axis_item)
    else:
        # Create mutable version of cube_like_item.
        try:
            cube_item_list = list(cube_like_item)
        except TypeError as err:
            if err.args[0] == "'NoneType' object is not iterable":
                cube_item_list = []
            else:
                raise err
        # Make sure cube_like_item is long enough to include common axis
        while len(cube_item_list) <= common_axis:
            cube_item_list.append(slice(None))
        # Create new sequence slice
        cube_item_list[common_axis] = sequence_slice.common_axis_item
        sequence_item = SequenceItem(sequence_slice.sequence_index, tuple(cube_item_list))
    return sequence_item


def convert_slice_nones_to_ints(slice_item, target_length):
    """
    Converts None types within a slice to the appropriate ints based on object to be sliced.

    The one case where a None is left in the slice object is when the step is negative and
    the stop parameter is None, since this scenario cannot be represented with an int stop
    parameter.

    Parameters
    ----------
    slice_item: `slice`
       Slice for which Nones should be converted.

    target_length: `int`
        Length of object to which slice will be applied.

    Returns
    -------
    new_slice: `slice`
        Slice with Nones replaced with ints.

    """
    if slice_item.step is None:
        step = 1
    else:
        step = slice_item.step
    start = slice_item.start
    stop = slice_item.stop
    if step < 0:
        if slice_item.start is None:
            start = int(target_length)
        stop = slice_item.stop
    else:
        if not slice_item.start:
            start = 0
        if not slice_item.stop:
            stop = int(target_length)
    return slice(start, stop, step)


def _get_axis_extra_coord_names_and_units(cube_list, axis):
    """
    Retrieve all extra coord names and units assigned to a data axis along a sequence of cubes.

    Parameters
    ----------
    cube_list: `list` of `ndcube.NDCube`
       The sequence of cubes from which to extract the extra coords.

    axis: `int`
       Number of axis (in data/numpy ordering convention).

    Returns
    -------
    axis_coord_names: `ndarray` of `str`
        Names of all extra coordinates in sequence.

    axis_coord_units: `ndarray` of `astropy.unit.unit`
        Units of extra coordinates.

    """
    # Define empty lists to hold results.
    axis_coord_names = []
    axis_coord_units = []
    # Extract extra coordinate names and units (if extra coord is a
    # quantity) from each cube.
    for cube in cube_list:
        all_extra_coords = cube.extra_coords
        if all_extra_coords is not None:
            all_extra_coords_keys = list(all_extra_coords.keys())
            for coord_key in all_extra_coords_keys:
                if all_extra_coords[coord_key]["axis"] == axis:
                    axis_coord_names.append(coord_key)
                    if isinstance(all_extra_coords[coord_key]["value"], u.Quantity):
                        axis_coord_units.append(all_extra_coords[coord_key]["value"].unit)
                    else:
                        axis_coord_units.append(None)
    # Extra coords common between cubes will be repeated.  Get rid of
    # duplicate names and then only keep the units corresponding to
    # the first occurence of that name.
    if len(axis_coord_names) > 0:
        axis_coord_names, ind = np.unique(np.asarray(axis_coord_names), return_index=True)
        axis_coord_units = np.asarray(axis_coord_units)[ind]
    else:
        axis_coord_names = None
        axis_coord_units = None
    return axis_coord_names, axis_coord_units


def _get_int_axis_extra_coords(cube_list, axis_coord_names, axis_coord_units, axis):
    """
    Retrieve all extra coord names and units assigned to a data axis along a sequence of cubes.

    Parameters
    ----------
    cube_list: `list` of `ndcube.NDCube`
       The sequence of cubes from which to extract the extra coords.

    axis_coord_names: `ndarray` of `str`
        Names of all extra coordinates in sequence.

    axis_coord_units: `ndarray` of `astropy.unit.unit`
        Units of extra coordinates.

    axis: `int`
       Number of axis (in data/numpy ordering convention).

    Returns
    -------
    axis_extra_coords: `dict`
        Extra coords along given axis.

    """
    # Define empty dictionary which will hold the extra coord
    # values not assigned a cube data axis.
    axis_extra_coords = {}
    # Iterate through cubes and populate values of each extra coord
    # not assigned a cube data axis.
    cube_extra_coords = [cube.extra_coords for cube in cube_list]
    for i, coord_key in enumerate(axis_coord_names):
        coord_values = []
        for j, cube in enumerate(cube_list):
            # Construct list of coord values from each cube for given extra coord.
            try:
                if isinstance(cube_extra_coords[j][coord_key]["value"], u.Quantity):
                    cube_coord_values = \
                      cube_extra_coords[j][coord_key]["value"].to(axis_coord_units[i]).value
                else:
                    cube_coord_values = np.asarray(cube_extra_coords[j][coord_key]["value"])
                coord_values = coord_values + list(cube_coord_values)
            except KeyError:
                # If coordinate not in cube, set coordinate values to NaN.
                coord_values = coord_values + [np.nan] * cube.dimensions[axis].value
        # Enter sequence extra coord into dictionary
        if axis_coord_units[i]:
            axis_extra_coords[coord_key] = coord_values * axis_coord_units[i]
        else:
            axis_extra_coords[coord_key] = np.asarray(coord_values)
    return axis_extra_coords



# -*- coding: utf-8 -*-

import abc
import warnings

import numpy as np
import astropy.nddata
import astropy.units as u
from astropy.utils.misc import InheritDocstrings

from ndcube import utils
from ndcube.ndcube_sequence import NDCubeSequence
from ndcube.utils.wcs import wcs_ivoa_mapping
from ndcube.mixins import NDCubeSlicingMixin, NDCubePlotMixin


__all__ = ['NDCubeABC', 'NDCubeBase', 'NDCube', 'NDCubeOrdered']


class NDCubeMetaClass(abc.ABCMeta, InheritDocstrings):
    """
    A metaclass that combines `abc.ABCMeta` and `~astropy.utils.misc.InheritDocstrings`.
    """


class NDCubeABC(astropy.nddata.NDData, metaclass=NDCubeMetaClass):

    @abc.abstractmethod
    def pixel_to_world(self, *quantity_axis_list):
        """
        Convert a pixel coordinate to a data (world) coordinate by using
        `~astropy.wcs.WCS.all_pix2world`.

        Parameters
        ----------
        quantity_axis_list : iterable
            An iterable of `~astropy.units.Quantity` with unit as pixel `pix`.
            Note that these quantities must be entered as separate arguments, not as one list.

        origin : `int`.
            Origin of the top-left corner. i.e. count from 0 or 1.
            Normally, origin should be 0 when passing numpy indices, or 1 if
            passing values from FITS header or map attributes.
            See `~astropy.wcs.WCS.wcs_pix2world` for more information.
            Default is 0.

        Returns
        -------
        coord : `list`
            A list of arrays containing the output coordinates
            reverse of the wcs axis order.
        """

    @abc.abstractmethod
    def world_to_pixel(self, *quantity_axis_list):
        """
        Convert a world coordinate to a data (pixel) coordinate by using
        `~astropy.wcs.WCS.all_world2pix`.

        Parameters
        ----------
        quantity_axis_list : iterable
            A iterable of `~astropy.units.Quantity`.
            Note that these quantities must be entered as separate arguments, not as one list.

        origin : `int`
            Origin of the top-left corner. i.e. count from 0 or 1.
            Normally, origin should be 0 when passing numpy indices, or 1 if
            passing values from FITS header or map attributes.
            See `~astropy.wcs.WCS.wcs_world2pix` for more information.
            Default is 0.

        Returns
        -------
        coord : `list`
            A list of arrays containing the output coordinates
            reverse of the wcs axis order.
        """

    # InheritDocstrings doesn't work on property methods.
    @abc.abstractproperty
    def dimensions(self):
        pass

    @abc.abstractproperty
    def world_axis_physical_types(self):
        pass

    @abc.abstractmethod
    def crop_by_coords(self, lower_corner, interval_widths=None, upper_corner=None, units=None):
        """
        Crops an NDCube given minimum values and interval widths along axes.

        Parameters
        ----------
        lower_corner: iterable of `astropy.units.Quantity` or `float`
            The minimum desired values along each relevant axis after cropping
            described in physical units consistent with the NDCube's wcs object.
            The length of the iterable must equal the number of data dimensions
            and must have the same order as the data.

        interval_widths: iterable of `astropy.units.Quantity` or `float`
            The width of the region of interest in each dimension in physical
            units consistent with the NDCube's wcs object. The length of the
            iterable must equal the number of data dimensions and must have
            the same order as the data. This argument will be removed in versions
            2.0, please use upper_corner argument.

        upper_corner: iterable of `astropy.units.Quantity` or `float`
            The maximum desired values along each relevant axis after cropping
            described in physical units consistent with the NDCube's wcs object.
            The length of the iterable must equal the number of data dimensions
            and must have the same order as the data.

        units: iterable of `astropy.units.quantity.Quantity`, optionnal
            If the inputs are set without units, the user must set the units
            inside this argument as `str`.
            The length of the iterable must equal the number of data dimensions
            and must have the same order as the data.

        Returns
        -------
        result: NDCube

        """


class NDCubeBase(NDCubeSlicingMixin, NDCubeABC):
    """
    Class representing N dimensional cubes.
    Extra arguments are passed on to `~astropy.nddata.NDData`.

    Parameters
    ----------
    data: `numpy.ndarray`
        The array holding the actual data in this object.

    wcs: `ndcube.wcs.wcs.WCS`
        The WCS object containing the axes' information

    uncertainty : any type, optional
        Uncertainty in the dataset. Should have an attribute uncertainty_type
        that defines what kind of uncertainty is stored, for example "std"
        for standard deviation or "var" for variance. A metaclass defining
        such an interface is NDUncertainty - but isn’t mandatory. If the uncertainty
        has no such attribute the uncertainty is stored as UnknownUncertainty.
        Defaults to None.

    mask : any type, optional
        Mask for the dataset. Masks should follow the numpy convention
        that valid data points are marked by False and invalid ones with True.
        Defaults to None.

    meta : dict-like object, optional
        Additional meta information about the dataset. If no meta is provided
        an empty collections.OrderedDict is created. Default is None.

    unit : Unit-like or str, optional
        Unit for the dataset. Strings that can be converted to a Unit are allowed.
        Default is None.

    extra_coords : iterable of `tuple`, each with three entries
        (`str`, `int`, `astropy.units.quantity` or array-like)
        Gives the name, axis of data, and values of coordinates of a data axis not
        included in the WCS object.

    copy : bool, optional
        Indicates whether to save the arguments as copy. True copies every attribute
        before saving it while False tries to save every parameter as reference.
        Note however that it is not always possible to save the input as reference.
        Default is False.

    missing_axis : `list` of `bool`
        Designates which axes in wcs object do not have a corresponding axis is the data.
        True means axis is "missing", False means axis corresponds to a data axis.
        Ordering corresponds to the axis ordering in the WCS object, i.e. reverse of data.
        For example, say the data's y-axis corresponds to latitude and x-axis corresponds
        to wavelength.  In order the convert the y-axis to latitude the WCS must contain
        a "missing" longitude axis as longitude and latitude are not separable.

    """

    def __init__(self, data, wcs, uncertainty=None, mask=None, meta=None,
                 unit=None, extra_coords=None, copy=False, missing_axis=None, **kwargs):
        if missing_axis is None:
            self.missing_axis = [False]*wcs.naxis
        else:
            self.missing_axis = missing_axis
        if data.ndim is not wcs.naxis:
            count = 0
            for bool_ in self.missing_axis:
                if not bool_:
                    count += 1
            if count is not data.ndim:
                raise ValueError("The number of data dimensions and number of "
                                 "wcs non-missing axes do not match.")
        # Format extra coords.
        if extra_coords:
            self._extra_coords_wcs_axis = \
              utils.cube._format_input_extra_coords_to_extra_coords_wcs_axis(
                  extra_coords, self.missing_axis, data.shape)
        else:
            self._extra_coords_wcs_axis = None
        # Initialize NDCube.
        super().__init__(data, wcs=wcs, uncertainty=uncertainty, mask=mask,
                         meta=meta, unit=unit, copy=copy, **kwargs)

    @property
    def dimensions(self):
        """
        Returns a named tuple with two attributes: 'shape' gives the shape
        of the data dimensions; 'axis_types' gives the WCS axis type of each dimension,
        e.g. WAVE or HPLT-TAN for wavelength of helioprojected latitude.

        """
        return u.Quantity(self.data.shape, unit=u.pix)

    @property
    def world_axis_physical_types(self):
        """
        Returns an iterable of strings describing the physical type for each world axis.

        The strings conform to the International Virtual Observatory Alliance
        standard, UCD1+ controlled Vocabulary.  For a description of the standard and
        definitions of the different strings and string components,
        see http://www.ivoa.net/documents/latest/UCDlist.html.

        """
        ctype = list(self.wcs.wcs.ctype)
        axes_ctype = []
        for i, axis in enumerate(self.missing_axis):
            if not axis:
                # Find keys in wcs_ivoa_mapping dict that represent start of CTYPE.
                # Ensure CTYPE is capitalized.
                keys = list(filter(lambda key: ctype[i].upper().startswith(key), wcs_ivoa_mapping))
                # Assuming CTYPE is supported by wcs_ivoa_mapping, use its corresponding axis name.
                if len(keys) == 1:
                    axis_name = wcs_ivoa_mapping.get(keys[0])
                # If CTYPE not supported, raise a warning and set the axis name to CTYPE.
                elif len(keys) == 0:
                    warnings.warn("CTYPE not recognized by ndcube. "
                                  "Please raise an issue at "
                                  "https://github.com/sunpy/ndcube/issues citing the "
                                  "unsupported CTYPE as we'll include it: "
                                  "CTYPE = {0}".format(ctype[i]))
                    axis_name = "custom:{0}".format(ctype[i])
                # If there are multiple valid keys, raise an error.
                else:
                    raise ValueError("Non-unique CTYPE key.  Please raise an issue at "
                                     "https://github.com/sunpy/ndcube/issues citing the "
                                     "following  CTYPE and non-unique keys: "
                                     "CTYPE = {0}; keys = {1}".format(ctype[i], keys))
                axes_ctype.append(axis_name)
        return tuple(axes_ctype[::-1])

    def pixel_to_world(self, *quantity_axis_list):
        # The docstring is defined in NDDataBase

        origin = 0
        list_arg = []
        indexed_not_as_one = []
        result = []
        quantity_index = 0
        for i in range(len(self.missing_axis)):
            wcs_index = self.wcs.naxis-1-i
            # the cases where the wcs dimension was made 1 and the missing_axis is True
            if self.missing_axis[wcs_index]:
                list_arg.append(self.wcs.wcs.crpix[wcs_index]-1+origin)
            else:
                # else it is not the case where the dimension of wcs is 1.
                list_arg.append(quantity_axis_list[quantity_index].to(u.pix).value)
                quantity_index += 1
                # appending all the indexes to be returned in the answer
                indexed_not_as_one.append(wcs_index)
        list_arguments = list_arg[::-1]
        pixel_to_world = self.wcs.all_pix2world(*list_arguments, origin)
        # collecting all the needed answer in this list.
        for index in indexed_not_as_one[::-1]:
            result.append(u.Quantity(pixel_to_world[index], unit=self.wcs.wcs.cunit[index]))
        return result[::-1]

    def world_to_pixel(self, *quantity_axis_list):
        # The docstring is defined in NDDataBase

        origin = 0
        list_arg = []
        indexed_not_as_one = []
        result = []
        quantity_index = 0
        for i in range(len(self.missing_axis)):
            wcs_index = self.wcs.naxis-1-i
            # the cases where the wcs dimension was made 1 and the missing_axis is True
            if self.missing_axis[wcs_index]:
                list_arg.append(self.wcs.wcs.crval[wcs_index]+1-origin)
            else:
                # else it is not the case where the dimension of wcs is 1.
                list_arg.append(
                    quantity_axis_list[quantity_index].to(self.wcs.wcs.cunit[wcs_index]).value)
                quantity_index += 1
                # appending all the indexes to be returned in the answer
                indexed_not_as_one.append(wcs_index)
        list_arguments = list_arg[::-1]
        world_to_pixel = self.wcs.all_world2pix(*list_arguments, origin)
        # collecting all the needed answer in this list.
        for index in indexed_not_as_one[::-1]:
            result.append(u.Quantity(world_to_pixel[index], unit=u.pix))
        return result[::-1]

    def axis_world_coords(self, *axes):
        """
        Returns WCS coordinate values of all pixels for all axes.

        Parameters
        ----------
        axes: `int` or `str`, or multiple `int` or `str`
            Axis number in numpy ordering or unique substring of
            `~ndcube.NDCube.world_axis_physical_types`
            of axes for which real world coordinates are desired.
            axes=None implies all axes will be returned.

        Returns
        -------
        axes_coords: `list` of `astropy.units.Quantity`
            Real world coords for axes in order requested by user.

        Example
        -------
        >>> NDCube.all_world_coords(('lat', 'lon')) # doctest: +SKIP
        >>> NDCube.all_world_coords(2) # doctest: +SKIP

        """
        # Define the dimensions of the cube and the total number of axes.
        cube_dimensions = np.array(self.dimensions.value, dtype=int)
        n_dimensions = cube_dimensions.size
        world_axis_types = self.world_axis_physical_types
        # Determine axis numbers of user supplied axes.
        if axes == ():
            int_axes = np.arange(n_dimensions)
        else:
            if isinstance(axes, int):
                int_axes = np.array([axes])
            elif isinstance(axes, str):
                int_axes = np.array([
                    utils.cube.get_axis_number_from_axis_name(axes, world_axis_types)])
            else:
                int_axes = np.empty(len(axes), dtype=int)
                for i, axis in enumerate(axes):
                    if isinstance(axis, int):
                        if axis < 0:
                            int_axes[i] = n_dimensions + axis
                        else:
                            int_axes[i] = axis
                    elif isinstance(axis, str):
                        int_axes[i] = utils.cube.get_axis_number_from_axis_name(
                            axis, world_axis_types)
        # Ensure user has not entered the same axis twice.
        repeats = set([x for x in int_axes if np.where(int_axes == x)[0].size > 1])
        if repeats:
            raise ValueError("The following axes were specified more than once: {}".format(
                ' '.join(map(str, repeats))))
        n_axes = len(int_axes)
        axes_coords = [None] * n_axes
        axes_translated = np.zeros_like(int_axes, dtype=bool)
        # Determine which axes are dependent on others.
        # Ensure the axes are in numerical order.
        dependent_axes = [list(utils.wcs.get_dependent_data_axes(self.wcs, axis, self.missing_axis))
                          for axis in int_axes]
        n_dependent_axes = [len(da) for da in dependent_axes]
        # Iterate through each axis and perform WCS translation.
        for i, axis in enumerate(int_axes):
            # If axis has already been translated, do not do so again.
            if not axes_translated[i]:
                if n_dependent_axes[i] == 1:
                    # Construct pixel quantities in each dimension letting
                    # other dimensions all have 0 pixel value.
                    quantity_list = [
                        u.Quantity(np.zeros(cube_dimensions[dependent_axes[i]]),
                                   unit=u.pix)] * n_dimensions
                    # Replace array in quantity list corresponding to current axis with
                    # np.arange array.
                    quantity_list[axis] = u.Quantity(np.arange(cube_dimensions[axis]), unit=u.pix)
                else:
                    # If the axis is dependent on another, perform
                    # translations on all dependent axes.
                    # Construct pixel quantities in each dimension letting
                    # other dimensions all have 0 pixel value.
                    quantity_list = [u.Quantity(np.zeros(tuple(
                        [cube_dimensions[k] for k in dependent_axes[i]])),
                        unit=u.pix)] * n_dimensions
                    # Construct orthogonal pixel index arrays for dependent axes.
                    dependent_pixel_quantities = np.meshgrid(
                        *[np.arange(cube_dimensions[k]) * u.pix
                          for k in dependent_axes[i]], indexing="ij")
                    for k, axis in enumerate(dependent_axes[i]):
                        quantity_list[axis] = dependent_pixel_quantities[k]
                # Perform wcs translation
                dependent_axes_coords = self.pixel_to_world(*quantity_list)
                # Place world coords into output list
                for dependent_axis in dependent_axes[i]:
                    if dependent_axis in int_axes:
                        # Due to error check above we know dependent
                        # axis can appear in int_axes at most once.
                        j = np.where(int_axes == dependent_axis)[0][0]
                        axes_coords[j] = dependent_axes_coords[dependent_axis]
                        # Remove axis from list that have now been translated.
                        axes_translated[j] = True
        if len(axes_coords) == 1:
            return axes_coords[0]
        else:
            return tuple(axes_coords)

    @property
    def extra_coords(self):
        """
        Dictionary of extra coords where each key is the name of an extra
        coordinate supplied by user during instantiation of the NDCube.

        The value of each key is itself a dictionary with the following keys:
          | 'axis': `int`
          |     The number of the data axis to which the extra coordinate corresponds.
          | 'value': `astropy.units.Quantity` or array-like
          |     The value of the extra coordinate at each pixel/array element along the
          |     corresponding axis (given by the 'axis' key, above).  Note this means
          |     that the length of 'value' must be equal to the length of the data axis
          |     to which is corresponds.
        """

        if not self._extra_coords_wcs_axis:
            result = None
        else:
            result = {}
            for key in list(self._extra_coords_wcs_axis.keys()):
                result[key] = {
                    "axis": utils.cube.wcs_axis_to_data_axis(
                        self._extra_coords_wcs_axis[key]["wcs axis"],
                        self.missing_axis),
                    "value": self._extra_coords_wcs_axis[key]["value"]}
        return result

    def crop_by_coords(self, lower_corner, interval_widths=None, upper_corner=None, units=None):
        # The docstring is defined in NDDataBase

        n_dim = self.data.ndim
        # Raising a value error if the arguments have not the same dimensions.
        # Calculation of upper_corner with the inputing interval_widths
        # This part of the code will be removed in version 2.0
        if interval_widths:
            warnings.warn(
                "interval_widths will be removed from the API in version 2.0, please use upper_corner argument.")
            if upper_corner:
                raise ValueError("Only one of interval_widths or upper_corner "
                                 "can be set. Recommend using upper_corner as "
                                 "interval_widths is deprecated.")
            if (len(lower_corner) != len(interval_widths)) or (len(lower_corner) != n_dim):
                raise ValueError("lower_corner and interval_widths must have "
                                 "same number of elements as number of data "
                                 "dimensions.")
            upper_corner = [lower_corner[i] + interval_widths[i] for i in range(n_dim)]
        # Raising a value error if the arguments have not the same dimensions.
        if (len(lower_corner) != len(upper_corner)) or (len(lower_corner) != n_dim):
            raise ValueError("lower_corner and upper_corner must have same"
                             "number of elements as number of data dimensions.")
        if units:
            # Raising a value error if units have not the data dimensions.
            if len(units) != n_dim:
                raise ValueError('units must have same number of elements as '
                                 'number of data dimensions.')
            # If inputs are not Quantity objects, they are modified into specified units
            lower_corner = [u.Quantity(lower_corner[i], unit=units[i])
                            for i in range(self.data.ndim)]
            upper_corner = [u.Quantity(upper_corner[i], unit=units[i])
                            for i in range(self.data.ndim)]
        else:
            if any([not isinstance(x, u.Quantity) for x in lower_corner + upper_corner]):
                raise TypeError("lower_corner and interval_widths/upper_corner must be "
                                "of type astropy.units.Quantity or the units kwarg "
                                "must be set.")
        # Get all corners of region of interest.
        all_world_corners_grid = np.meshgrid(
            *[u.Quantity([lower_corner[i], upper_corner[i]], unit=lower_corner[i].unit).value
              for i in range(self.data.ndim)])
        all_world_corners = [all_world_corners_grid[i].flatten()*lower_corner[i].unit
                             for i in range(n_dim)]
        # Convert to pixel coordinates
        all_pix_corners = self.world_to_pixel(*all_world_corners)
        # Derive slicing item with which to slice NDCube.
        # Be sure to round down min pixel and round up + 1 the max pixel.
        item = tuple([slice(int(np.clip(axis_pixels.value.min(), 0, None)),
                            int(np.ceil(axis_pixels.value.max()))+1)
                      for axis_pixels in all_pix_corners])
        return self[item]

    def crop_by_extra_coord(self, coord_name, min_coord_value, max_coord_value):
        """
        Crops an NDCube given a minimum value and interval width along an extra coord.

        Parameters
        ----------
        coord_name: `str`
            Name of extra coordinate by which to crop.

        min_coord_value: Single value `astropy.units.Quantity`
            The minimum desired value of the extra coord after cropping.
            Unit must be consistent with the extra coord on which cropping is based.

        min_coord_value: Single value `astropy.units.Quantity`
            The maximum desired value of the extra coord after cropping.
            Unit must be consistent with the extra coord on which cropping is based.

        Returns
        -------
        result: `ndcube.NDCube`

        """
        if not isinstance(coord_name, str):
            raise TypeError("The API for this function has changed. "
                            "Please give coord_name, min_coord_value, max_coord_value")
        extra_coord_dict = self.extra_coords[coord_name]
        if isinstance(extra_coord_dict["value"], u.Quantity):
            extra_coord_values = extra_coord_dict["value"]
        else:
            extra_coord_values = np.asarray(extra_coord_dict["value"])
        w = np.logical_and(extra_coord_values >= min_coord_value,
                           extra_coord_values < max_coord_value)
        w = np.arange(len(extra_coord_values))[w]
        item = [slice(None)]*len(self.dimensions)
        item[extra_coord_dict["axis"]] = slice(w[0], w[-1]+1)
        return self[tuple(item)]

    def __repr__(self):
        return (
            """NDCube
---------------------
{wcs}
---------------------
Length of NDCube: {lengthNDCube}
Axis Types of NDCube: {axis_type}
""".format(wcs=self.wcs.__repr__(), lengthNDCube=self.dimensions,
           axis_type=self.world_axis_physical_types))

    def explode_along_axis(self, axis):
        """
        Separates slices of NDCubes along a given cube axis into a NDCubeSequence
        of (N-1)DCubes.

        Parameters
        ----------
        axis : `int`
            The axis along which the data is to be changed.

        Returns
        -------
        result : `ndcube_sequence.NDCubeSequence`

        """
        # If axis is -ve then calculate the axis from the length of the dimensions of one cube
        if axis < 0:
            axis = len(self.dimensions) + axis
        # To store the resultant cube
        result_cubes = []
        # All slices are initially initialised as slice(None, None, None)
        cube_slices = [slice(None, None, None)] * self.data.ndim
        # Slicing the cube inside result_cube
        for i in range(self.data.shape[axis]):
            # Setting the slice value to the index so that the slices are done correctly.
            cube_slices[axis] = i
            # Set to None the metadata of sliced cubes.
            item = tuple(cube_slices)
            sliced_cube = self[item]
            sliced_cube.meta = None
            # Appending the sliced cubes in the result_cube list
            result_cubes.append(sliced_cube)
        # Creating a new NDCubeSequence with the result_cubes and common axis as axis
        return NDCubeSequence(result_cubes, common_axis=axis, meta=self.meta)

class NDCube(NDCubeBase, NDCubePlotMixin, astropy.nddata.NDArithmeticMixin):
    pass


class NDCubeOrdered(NDCube):
    """
    Class representing N dimensional cubes with oriented WCS.
    Extra arguments are passed on to NDData's init.
    The order is TIME, SPECTRAL, SOLAR-x, SOLAR-Y and any other dimension.
    For example, in an x, y, t cube the order would be (t,x,y) and in a
    lambda, t, y cube the order will be (t, lambda, y).
    Extra arguments are passed on to NDData's init.

    Parameters
    ----------
    data: `numpy.ndarray`
        The array holding the actual data in this object.

    wcs: `ndcube.wcs.wcs.WCS`
        The WCS object containing the axes' information. The axes'
        priorities are time, spectral, celestial. This means that if
        present, each of these axis will take precedence over the others.

    uncertainty : any type, optional
        Uncertainty in the dataset. Should have an attribute uncertainty_type
        that defines what kind of uncertainty is stored, for example "std"
        for standard deviation or "var" for variance. A metaclass defining
        such an interface is NDUncertainty - but isn’t mandatory. If the uncertainty
        has no such attribute the uncertainty is stored as UnknownUncertainty.
        Defaults to None.

    mask : any type, optional
        Mask for the dataset. Masks should follow the numpy convention
        that valid data points are marked by False and invalid ones with True.
        Defaults to None.

    meta : dict-like object, optional
        Additional meta information about the dataset. If no meta is provided
        an empty collections.OrderedDict is created. Default is None.

    unit : Unit-like or str, optional
        Unit for the dataset. Strings that can be converted to a Unit are allowed.
        Default is None.

    copy : bool, optional
        Indicates whether to save the arguments as copy. True copies every attribute
        before saving it while False tries to save every parameter as reference.
        Note however that it is not always possible to save the input as reference.
        Default is False.
    """

    def __init__(self, data, wcs, uncertainty=None, mask=None, meta=None,
                 unit=None, extra_coords=None, copy=False, missing_axis=None, **kwargs):
        axtypes = list(wcs.wcs.ctype)[::-1]
        array_order = utils.cube.select_order(axtypes)
        result_data = data.transpose(array_order)
        result_wcs = utils.wcs.reindex_wcs(wcs, np.array(array_order))
        if uncertainty is not None:
            result_uncertainty = uncertainty.transpose(array_order)
        else:
            result_uncertainty = None
        if mask is not None:
            result_mask = mask.transpose(array_order)
        else:
            result_mask = None
        # Reorder extra coords if needed.
        if extra_coords:
            reordered_extra_coords = []
            for coord in extra_coords:
                coord_list = list(coord)
                coord_list[1] = array_order[coord_list[1]]
                reordered_extra_coords.append(tuple(coord_list))

        super().__init__(result_data, result_wcs, uncertainty=result_uncertainty,
                         mask=result_mask, meta=meta, unit=unit,
                         extra_coords=reordered_extra_coords,
                         copy=copy, missing_axis=missing_axis, **kwargs)
